# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

class Builder(object):
  """Abstract base class for builder implementations that can execute a parsed BUILD target."""

  def __init__(self, ferror, root_dir):
    self.ferror = ferror
    self.root_dir = root_dir

  def build(self, target, args):
    """Subclasses must implement a BUILD target executor.  The value returned should be an int,
    0 indicating success and any other value indicating failure.

    target: the parsed target to build
    args: additional arguments to the builder backend"""

    pass



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import os
import sys

from twitter.common import dirutil, log
from twitter.common.process import spawn_daemon
from twitter.common.quantity.parse_simple import parse_time
from twitter.pants import get_buildroot
from twitter.pants.buildtimestats import StatsUploader

STATS_COLLECTION_URL = "devprod_stats.production.devprod.service.smf1.twitter.com"
STATS_COLLECTION_PORT = 80
STATS_COLLECTION_ENDPOINT = "/buildtime_stats.json"
MAX_UPLOAD_DELAY = "6h"
DEFAULT_STATS_FILE = ".pants.stats"
PHASE_TOTAL = "phase_total"
CMD_TOTAL = "cmd_total"

__author__ = 'Tejal Desai'


class BuildTimeStats(object):

  def __init__(self, user, force_upload):
    """
    :param user: The user to be used for uploading stats. Its the current user.
    :param force_upload: [False] Uploads Stats at the end of pants run. 
    """
    self._user = user
    self._force_upload = force_upload
    self._max_delay = parse_time(MAX_UPLOAD_DELAY)

  def _get_default_stats_file(self):
    return os.path.join(get_buildroot(), DEFAULT_STATS_FILE)

  def compute_stats(self, executed_goals, elapsed):
    timings_array = []
    for phase, timings in executed_goals.items():
      phase_time = None
      for goal, times in timings.items():
        #Create a new structure
        timing = dict()
        timing['phase'] = str(phase)
        timing['goal'] = goal
        timing['total'] = sum(times)
        if not phase_time:
          phase_time = 0
        phase_time += sum(times)
        #Add the timings for each sub phase in the timings array
        timings_array.append(timing)
      if len(timings) > 1:
        #Add the phase total
        timing = dict()
        timing['phase'] = str(phase)
        timing['goal'] = PHASE_TOTAL
        timing['total'] = phase_time
        timings_array.append(timing)
    #end of Loop through PHASES
    timing = {}
    timing['phase'] = CMD_TOTAL
    timing['goal'] = CMD_TOTAL
    timing['total'] = elapsed
    timings_array.append(timing)
    return timings_array

  def stats_uploader_daemon(self, stats):
    """
    Starts the StatsUploader as a daemon process if it is already not running
    """
    log.debug("Checking if the statsUploaderDaemon is already running")
    stats_pid = os.path.join("/tmp", self._user, ".pid_stats")
    stats_uploader_dir = os.path.join("/tmp", self._user)
    dirutil.safe_mkdir(stats_uploader_dir)
    if not os.path.exists(stats_pid):
      log.debug("Starting the daemon")
      stats_log_file = os.path.join("/tmp", self._user, "buildtime_uploader")
      log.debug("The logs are written to %s" % stats_log_file)
      if spawn_daemon(pidfile=stats_pid, quiet=True):
        su = StatsUploader(STATS_COLLECTION_URL, STATS_COLLECTION_PORT, STATS_COLLECTION_ENDPOINT,
                           self._max_delay, self._get_default_stats_file(), self._user,
                           self._force_upload)
        su.upload_sync(stats)

  def record_stats(self, timings, elapsed):
    """Records all the stats for -x flag
    and the network stats
    """
    timing_stats = self.compute_stats(timings, elapsed)
    stats = {}
    stats["timings"] = timing_stats
    stats["args"] = sys.argv[1:]
    self.stats_uploader_daemon(stats)



# ==================================================================================================
# Copyright 2014 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.pants.base.build_manual import manual
from twitter.pants.targets.exclude import Exclude
from twitter.pants.targets.jvm_target import JvmTarget
from twitter.pants.targets.hadoop_binary import ON_HADOOP_CLUSTER_EXCLUDES

OINK_EXCLUDES = ON_HADOOP_CLUSTER_EXCLUDES + [
  Exclude('org.apache.pig'),
  Exclude('org.apache.hbase', 'hbase')
]

@manual.builddict(tags=["jvm"])
class OinkQuery(JvmTarget):
  """Represents an OinkQuery target.

  Goals such as ``bundle`` will create a OinkQuery bundle.
  """

  def __init__(self, name, dependencies, sources=None, excludes=None, deploy_excludes=None):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param dependencies: List of targets (probably ``java_library`` targets
      to "link" in.
    :param string sources: Optional list of ``.yml`` files that represent the
      Oink Query job.
    :param excludes: List of ``exclude``\s to filter this target's transitive
      dependencies against.
    :param deploy_excludes: List of ``excludes`` to apply at deploy time.
      If you, for example, deploy a a servlet that has one version of
      ``servlet.jar`` onto a Tomcat environment that provides another version,
      they might conflict. ``deploy_excludes`` gives you a way to build your
      code but exclude the conflicting ``jar`` when deploying.
    """
    JvmTarget.__init__(self, name, sources, dependencies, excludes)
    # TODO: configurations is required when fetching jar_dependencies but should not be
    self.configurations = None
    self.deploy_excludes = (deploy_excludes or []) + OINK_EXCLUDES



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.common.collections import OrderedSet

from twitter.pants.base import ParseContext, Target
from twitter.pants.base.target import TargetDefinitionException
from twitter.pants.targets import PythonRequirement


def is_python_root(target):
  return isinstance(target, PythonRoot)


class PythonRoot(Target):
  """
    Internal target for managing python chroot state.
  """
  @classmethod
  def synthetic_name(cls, targets):
    return list(targets)[0].name if len(targets) > 0 else 'empty'

  @classmethod
  def union(cls, targets, name=None):
    name = name or (cls.synthetic_name(targets) + '-union')
    with ParseContext.temp():
      return cls(name, dependencies=targets)

  @classmethod
  def of(cls, target):
    with ParseContext.temp():
      return cls(target.name, dependencies=[target])

  def __init__(self, name, dependencies=None):
    self.dependencies = OrderedSet(dependencies) if dependencies else OrderedSet()
    self.internal_dependencies = OrderedSet()
    self.interpreters = []
    self.distributions = {} # interpreter => distributions
    self.chroots = {}       # interpreter => chroots
    super(PythonRoot, self).__init__(name)

  def closure(self):
    os = OrderedSet()
    for target in self.dependencies | self.internal_dependencies:
      os.update(target.closure())
    return os

  def select(self, target_class):
    return OrderedSet(target for target in self.closure() if isinstance(target, target_class))

  @property
  def requirements(self):
    return self.select(PythonRequirement)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import os
import sys


from twitter.common import log

from .build_root import BuildRoot
from .version import VERSION as _VERSION


def get_version():
  return _VERSION


def get_buildroot():
  """Returns the pants ROOT_DIR, calculating it if needed."""
  try:
    return BuildRoot().path
  except BuildRoot.NotFoundError as e:
    print(e.message, file=sys.stderr)
    sys.exit(1)


from twitter.pants.scm import Scm

_SCM = None


def get_scm():
  """Returns the pants Scm if any."""
  # TODO(John Sirois): Extract a module/class to carry the bootstrap logic.
  global _SCM
  if not _SCM:
    # We know about git, so attempt an auto-configure
    git_dir = os.path.join(get_buildroot(), '.git')
    if os.path.isdir(git_dir):
      from twitter.pants.scm.git import Git
      git = Git(worktree=get_buildroot())
      try:
        log.info('Detected git repository on branch %s' % git.branch_name)
        set_scm(git)
      except git.LocalException:
        pass
  return _SCM


def set_scm(scm):
  """Sets the pants Scm."""
  if scm is not None:
    if not isinstance(scm, Scm):
      raise ValueError('The scm must be an instance of Scm, given %s' % scm)
    global _SCM
    _SCM = scm


from twitter.common.dirutil import Fileset

globs = Fileset.globs
rglobs = Fileset.rglobs


def is_concrete(target):
  """Returns true if a target resolves to itself."""
  targets = list(target.resolve())
  return len(targets) == 1 and targets[0] == target


from twitter.pants.targets import *

# aliases
annotation_processor = AnnotationProcessor
artifact = Artifact
benchmark = Benchmark
bundle = Bundle
credentials = Credentials
dependencies = jar_library = JarLibrary
egg = PythonEgg
exclude = Exclude
fancy_pants = Pants
gem = Gem
hadoop_binary = TwitterHadoopBinary
idl_jar_thrift_library = IdlJvmThriftLibrary
jar = JarDependency
java_library = JavaLibrary
java_antlr_library = JavaAntlrLibrary
java_protobuf_library = JavaProtobufLibrary
junit_tests = java_tests = JavaTests
java_thrift_library = JavaThriftLibrary
# TODO(Anand) Remove this from pants proper when a code adjoinment mechanism exists
# or ok if/when thriftstore is open sourced as well
java_thriftstore_dml_library = JavaThriftstoreDMLLibrary
jvm_binary = JvmBinary
jvm_app = JvmApp
oink_query = OinkQuery
page = Page
python_artifact = setup_py = PythonArtifact
python_binary = PythonBinary
python_library = PythonLibrary
python_antlr_library = PythonAntlrLibrary
python_requirement = PythonRequirement
python_thrift_library = PythonThriftLibrary
python_tests = PythonTests
python_test_suite = PythonTestSuite
repo = Repository
resources = Resources
ruby_thrift_library = RubyThriftLibrary
scala_library = ScalaLibrary
scala_specs = scala_tests = ScalaTests
scalac_plugin = ScalacPlugin
source_root = SourceRoot
storm_binary = TwitterStormBinary
thrift_jar = ThriftJar
thrift_library = ThriftLibrary
wiki = Wiki


def is_codegen(target):
  """Returns True if the target is synthesized from generated sources or represents a set of idl
  files that generate sources."""

  return hasattr(target, 'is_codegen') and target.is_codegen


def has_sources(target, extension=None):
  """Returns True if the target has sources.

  If an extension is supplied the target is further checked for at least 1 source with the given
  extension. It also excludes Resources when extension is provided.
  """
  return (isinstance(target, TargetWithSources)
          and (not extension
               or (not isinstance(target, Resources)
                   and any(filter(lambda source: source.endswith(extension), target.sources)))))


def has_resources(target):
  """Returns True if the target has an associated set of Resources."""
  return hasattr(target, 'resources') and target.resources


def is_exported(target):
  """Returns True if the target provides an artifact exportable from the repo."""
  return hasattr(target, 'provides') and target.provides


def is_internal(target):
  """Returns True if the target is internal to the repo (ie: it might have dependencies)."""
  return isinstance(target, InternalTarget)


def is_jar(target):
  """Returns True if the target is a jar."""
  return isinstance(target, JarDependency)


def is_jvm(target):
  """Returns True if the target produces jvm bytecode."""
  return isinstance(target, JvmTarget)


def has_jvm_targets(targets):
  """Returns true if the given sequence of targets contains at least one jvm target as determined
  by is_jvm(...)"""

  return len(list(extract_jvm_targets(targets))) > 0


def extract_jvm_targets(targets):
  """Returns an iterator over the jvm targets the given sequence of targets resolve to.  The given
  targets can be a mix of types and only valid jvm targets (as determined by is_jvm(...) will be
  returned by the iterator."""

  for target in targets:
    if target is None:
      print('Warning! Null target!', file=sys.stderr)
      continue
    for real_target in target.resolve():
      if is_jvm(real_target):
        yield real_target


def is_java(target):
  """Returns True if the target has or generates java sources."""
  return (isinstance(target, (AnnotationProcessor, JavaLibrary, JavaProtobufLibrary, JavaTests))
          or is_thrift(target))


def is_jvm_app(target):
  """Returns True if the target produces a java application with bundled auxiliary files."""
  return isinstance(target, JvmApp)


def is_thrift(target):
  """Returns True if the target has thrift IDL sources."""
  return isinstance(target, JavaThriftLibrary)


def is_apt(target):
  """Returns True if the target produces an annotation processor."""
  return isinstance(target, AnnotationProcessor)


def is_python(target):
  """Returns True if the target has python sources."""
  return isinstance(target, PythonTarget) or isinstance(target, PythonRequirement)


def is_scala(target):
  """Returns True if the target has scala sources."""
  return isinstance(target, (ScalaLibrary, ScalaTests))


def is_scalac_plugin(target):
  """Returns True if the target defines a scalac plugin."""
  return isinstance(target, ScalacPlugin)


def is_test(t):
  """Returns True if the target is comprised of tests."""
  return isinstance(t, (JavaTests, ScalaTests, PythonTests))


def maven_layout(basedir=None):
  """Sets up typical maven project source roots for all built-in pants target types.

  Shortcut for ``source_root('src/main/java', *java targets*)``,
  ``source_root('src/main/python', *python targets*)``, ...

  :param string basedir: Instead of using this BUILD file's directory as
    the base of the source tree, use a subdirectory. E.g., instead of
    expecting to find java files in ``src/main/java``, expect them in
    ``**basedir**/src/main/java``.
  """

  def root(path, *types):
    source_root(os.path.join(basedir, path) if basedir else path, *types)

  root('src/main/antlr', java_antlr_library, page, python_antlr_library)
  root('src/main/java', annotation_processor, java_library, jvm_binary, page)
  root('src/main/protobuf', java_protobuf_library, page)
  root('src/main/python', page, python_binary, python_library)
  root('src/main/resources', page, resources)
  root('src/main/scala', jvm_binary, page, scala_library)
  root('src/main/thrift', java_thrift_library, page, python_thrift_library, ruby_thrift_library, thrift_library)

  root('src/test/java', java_library, junit_tests, page)
  root('src/test/python', page, python_library, python_tests, python_test_suite)
  root('src/test/resources', page, resources)
  root('src/test/scala', junit_tests, page, scala_library, scala_specs)


# bind this as late as possible
pants = fancy_pants

# bind tasks and goals below utility functions they use from above
from twitter.pants.base import Config
from twitter.pants.goal import Context, Goal, Group, Phase
from twitter.pants.tasks import Task, TaskError

goal = Goal
group = Group
phase = Phase


# TODO(John Sirois): Update to dynamic linking when http://jira.local.twitter.com/browse/AWESOME-243
# is avaiable.
# bind twitter-specific idl helper
from twitter.pants.tasks.extract import Extract

compiled_idl = Extract.compiled_idl


# All these exports are visible in bootstrap_buildfiles
__all__ = (
  'annotation_processor',
  'artifact',
  'benchmark',
  'bundle',
  'compiled_idl',
  'credentials',
  'dependencies',
  'exclude',
  'egg',
  'gem',
  'get_buildroot',
  'get_scm',
  'get_version',
  'globs',
  'goal',
  'group',
  'hadoop_binary',
  'idl_jar_thrift_library',
  'is_apt',
  'is_codegen',
  'is_exported',
  'is_internal',
  'is_jar',
  'is_java',
  'is_jvm',
  'is_python',
  'is_scala',
  'is_test',
  'jar',
  'jar_library',
  'java_antlr_library',
  'java_library',
  'java_protobuf_library',
  'java_tests',
  'java_thrift_library',
  'java_thriftstore_dml_library',
  'junit_tests',
  'jvm_app',
  'jvm_binary',
  'maven_layout',
  'oink_query',
  'page',
  'pants',
  'phase',
  'python_antlr_library',
  'python_artifact',
  'python_binary',
  'python_library',
  'python_requirement',
  'python_tests',
  'python_test_suite',
  'python_thrift_library',
  'repo',
  'resources',
  'rglobs',
  'ruby_thrift_library',
  'scala_library',
  'scala_specs',
  'scala_tests',
  'scalac_plugin',
  'setup_py',
  'source_root',
  'storm_binary',
  'thrift_jar',
  'thrift_library',
  'wiki',
  'Config',
  'Context',
  'JavaLibrary',
  'JavaTests',
  'Task',
  'TaskError'
)



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.common.collections import OrderedSet
from twitter.pants.base import Target


class TargetWithDependencies(Target):
  def __init__(self, name, dependencies=None):
    Target.__init__(self, name)
    self.dependencies = OrderedSet(dependencies) if dependencies else OrderedSet()

  def _walk(self, walked, work, predicate=None):
    Target._walk(self, walked, work, predicate)
    for dependency in self.dependencies:
      for dep in dependency.resolve():
        if isinstance(dep, Target) and not dep in walked:
          walked.add(dep)
          if not predicate or predicate(dep):
            additional_targets = work(dep)
            dep._walk(walked, work, predicate)
            if additional_targets:
              for additional_target in additional_targets:
                if hasattr(additional_target, '_walk'):
                  additional_target._walk(walked, work, predicate)


# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import json

class ThriftJSONEncoder(json.JSONEncoder):
  """An encoder that makes python thrift structs JSON serializable via the
  standard python json module.

  Pass this encoder when writing json, like this:

  json.dumps(thriftobj, cls=text.ThriftJSONEncoder, <other kwargs>)

  Note that this is not a full protocol implementation in the thrift sense. This
  is just a quick-and-easy pretty-printer for unittests, debugging etc.
  """
  THRIFT_SPEC = 'thrift_spec'

  def __init__(self, *args, **kwargs):
    super(ThriftJSONEncoder, self).__init__(*args, **kwargs)

  def default(self, o):
    # Handle sets (the value type of a thrift set field). We emit them as lists,
    # sorted by element.
    if isinstance(o, set):
      ret = list(o)
      ret.sort()
      return ret

    # Handle everything that isn't a thrift struct.
    if not hasattr(o, ThriftJSONEncoder.THRIFT_SPEC):
      return super(ThriftJSONEncoder, self).default(o)

    # Handle thrift structs.
    spec = getattr(o, ThriftJSONEncoder.THRIFT_SPEC)
    ret = {}
    for (field_number, type, name, type_info, default) in \
          [field_spec for field_spec in spec if field_spec is not None]:
      if name in o.__dict__:
        val = o.__dict__[name]
        if val != default:
          ret[name] = val
    return ret

def thrift_to_json(o):
  """A utility shortcut function to return a pretty-printed JSON thrift object.

  Map keys will be sorted, making the returned string suitable for use in test comparisons."""
  # ident=2 tells python to pretty-print, and put a newline after each comma. Therefore we
  # set the comma separator to have no space after it. Otherwise it'll be difficult to embed
  # a golden string for comparisons (since our editors strip spaces at the ends of lines).
  return json.dumps(o, sort_keys=True, cls=ThriftJSONEncoder, indent=2, separators=(',', ': '))




# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

__author__ = 'Brian Larson'

import os

from twitter.common.collections import OrderedSet
from twitter.common.dirutil import safe_mkdir

from twitter.pants import is_jvm
from twitter.pants.targets import JavaLibrary, JavaAntlrLibrary
from twitter.pants.tasks import TaskError
from twitter.pants.tasks.code_gen import CodeGen
from twitter.pants.tasks.nailgun_task import NailgunTask


class AntlrGen(CodeGen, NailgunTask):

  # Maps the compiler attribute of a target to the config key in pants.ini
  _CONFIG_SECTION_BY_COMPILER = {
    'antlr3': 'antlr-gen',
    'antlr4': 'antlr4-gen',
  }

  def __init__(self, context):
    CodeGen.__init__(self, context)
    NailgunTask.__init__(self, context)
    self._classpath_by_profile = {}

  def is_gentarget(self, target):
    return isinstance(target, JavaAntlrLibrary)

  def is_forced(self, lang):
    return True

  def genlangs(self):
    return dict(java=is_jvm)

  def prepare_gen(self, targets):
    for target in targets:
      antlr_profile = self._antlr_profile(target)
      if antlr_profile not in self._classpath_by_profile:
        antlr_classpath = self.profile_classpath(antlr_profile)
        self._classpath_by_profile[antlr_profile] = antlr_classpath

  def genlang(self, lang, targets):
    if lang != 'java':
      raise TaskError('Unrecognized antlr gen lang: %s' % lang)

    # TODO: Instead of running the compiler for each target, collect the targets
    # by type and invoke it twice, once for antlr3 and once for antlr4.

    for target in targets:
      java_out = self._java_out(target)
      safe_mkdir(java_out)

      antlr_profile = self._antlr_profile(target)
      antlr_classpath = self._classpath_by_profile[antlr_profile]
      args = ["-o", java_out]

      if target.compiler == 'antlr3':
        antlr_main = 'org.antlr.Tool'
      elif target.compiler == 'antlr4':
        args.append("-visitor")  # Generate Parse Tree Visitor As Well
        antlr_main = 'org.antlr.v4.Tool'
      else:
        raise TaskError("Unknown ANTLR compiler: {}".format(target.compiler))

      sources = self._calculate_sources([target])
      args.extend(sources)
      result = self.runjava(antlr_classpath, antlr_main, args=args)
      if result != 0:
        raise TaskError('java %s ... exited non-zero (%i)' % (antlr_main, result))

  def _calculate_sources(self, targets):
    sources = set()

    def collect_sources(target):
      if self.is_gentarget(target):
        sources.update(os.path.join(target.target_base, source) for source in target.sources)
    for target in targets:
      target.walk(collect_sources)
    return sources

  def createtarget(self, lang, gentarget, dependees):
    if lang != 'java':
      raise TaskError('Unrecognized antlr gen lang: %s' % lang)
    return self._create_java_target(gentarget, dependees)

  def _create_java_target(self, target, dependees):
    antlr_files_suffix = ["Lexer.java", "Parser.java"]
    if (target.compiler == 'antlr4'):
      antlr_files_suffix = ["BaseListener.java", "BaseVisitor.java",
                            "Listener.java", "Visitor.java"] + antlr_files_suffix

    generated_sources = []
    for source in target.sources:
      # Antlr enforces that generated sources are relative to the base filename, and that
      # each grammar filename must match the resulting grammar Lexer and Parser classes.
      source_base, source_ext = os.path.splitext(source)
      for suffix in antlr_files_suffix:
        generated_sources.append(source_base + suffix)

    deps = self._resolve_java_deps(target)

    tgt = self.context.add_new_target(os.path.join(self._java_out(target), target.target_base),
                                      JavaLibrary,
                                      name=target.id,
                                      provides=target.provides,
                                      sources=generated_sources,
                                      dependencies=deps)
    tgt.id = target.id
    tgt.is_codegen = True
    for dependee in dependees:
      dependee.update_dependencies([tgt])
    return tgt

  def _resolve_java_deps(self, target):
    key = self._CONFIG_SECTION_BY_COMPILER[target.compiler]

    deps = OrderedSet()
    for dep in self.context.config.getlist(key, 'javadeps'):
        deps.update(self.context.resolve(dep))
    return deps

  def _antlr_profile(self, target):
    key = self._CONFIG_SECTION_BY_COMPILER[target.compiler]
    return self.context.config.get(key, 'antlr_profile')

  def _java_out(self, target):
    key = self._CONFIG_SECTION_BY_COMPILER[target.compiler]
    return os.path.join(self.context.config.get(key, 'workdir'), 'gen-java')



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.pants.targets.exportable_jvm_library import ExportableJvmLibrary
from twitter.pants.targets.resources import WithLegacyResources
from twitter.pants.base import manual


@manual.builddict(tags=["java"])
class AnnotationProcessor(ExportableJvmLibrary, WithLegacyResources):
  """Produces a Java library containing one or more annotation processors."""

  def __init__(self, name, sources,
               provides = None,
               dependencies = None,
               excludes = None,
               resources = None,
               processors = None):

    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param sources: A list of filenames representing the source code
      this library is compiled from.
    :type sources: list of strings
    :param Artifact provides:
      The :class:`twitter.pants.targets.artifact.Artifact`
      to publish that represents this target outside the repo.
    :param dependencies: List of :class:`twitter.pants.base.target.Target` instances
      this target depends on.
    :type dependencies: list of targets
    :param excludes: List of :class:`twitter.pants.targets.exclude.Exclude` instances
      to filter this target's transitive dependencies against.
    :param resources: An optional list of file paths (DEPRECATED) or
      ``resources`` targets (which in turn point to file paths). The paths
      indicate text file resources to place in this module's jar.
    :param processors: A list of the fully qualified class names of the
      annotation processors this library exports.
    """

    ExportableJvmLibrary.__init__(self,
                                  name,
                                  sources,
                                  provides,
                                  dependencies,
                                  excludes,
                                  ())

    WithLegacyResources.__init__(self, name, sources=sources, resources=resources)

    self.processors = processors



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================
import os
import pkgutil

from collections import defaultdict

from twitter.common.collections import OrderedSet
from twitter.common.dirutil import safe_delete, safe_mkdir, safe_open

from twitter.pants import get_buildroot
from twitter.pants.base.generator import TemplateData, Generator
from twitter.pants.tasks.ide_gen import IdeGen

__author__ = 'John Sirois'


_TEMPLATE_BASEDIR = os.path.join('templates', 'eclipse')


_VERSIONS = {
  '3.5': '3.7', # 3.5-3.7 are .project/.classpath compatible
  '3.6': '3.7',
  '3.7': '3.7',
}


_SETTINGS = (
  'org.eclipse.core.resources.prefs',
  'org.eclipse.jdt.ui.prefs',
)


class EclipseGen(IdeGen):
  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    IdeGen.setup_parser(option_group, args, mkflag)

    supported_versions = sorted(list(_VERSIONS.keys()))
    option_group.add_option(mkflag("eclipse-version"), dest="eclipse_gen_version",
                            default='3.6', type="choice", choices=supported_versions,
                            help="[%%default] The Eclipse version the project "
                                   "configuration should be generated for; can be one of: "
                                   "%s" % supported_versions)

  def __init__(self, context):
    IdeGen.__init__(self, context)

    eclipse_version = _VERSIONS[context.options.eclipse_gen_version]
    self.project_template = os.path.join(_TEMPLATE_BASEDIR, 'project-%s.mk' % eclipse_version)
    self.classpath_template = os.path.join(_TEMPLATE_BASEDIR, 'classpath-%s.mk' % eclipse_version)
    self.apt_template = os.path.join(_TEMPLATE_BASEDIR, 'factorypath-%s.mk' % eclipse_version)
    self.pydev_template = os.path.join(_TEMPLATE_BASEDIR, 'pydevproject-%s.mk' % eclipse_version)
    self.debug_template = os.path.join(_TEMPLATE_BASEDIR, 'debug-launcher-%s.mk' % eclipse_version)
    self.coreprefs_template = os.path.join(_TEMPLATE_BASEDIR,
                                           'org.eclipse.jdt.core.prefs-%s.mk' % eclipse_version)

    self.project_filename = os.path.join(self.cwd, '.project')
    self.classpath_filename = os.path.join(self.cwd, '.classpath')
    self.apt_filename = os.path.join(self.cwd, '.factorypath')
    self.pydev_filename = os.path.join(self.cwd, '.pydevproject')
    self.coreprefs_filename = os.path.join(self.cwd, '.settings', 'org.eclipse.jdt.core.prefs')

  def generate_project(self, project):
    def linked_folder_id(path):
      return path.replace(os.path.sep, '.')

    def base_path(source_set):
      return os.path.join(source_set.root_dir, source_set.source_base)

    source_bases = {}
    for source_set in project.sources:
      source_bases[base_path(source_set)] = linked_folder_id(source_set.source_base)
    if project.has_python:
      for source_set in project.py_sources:
        source_bases[base_path(source_set)] = linked_folder_id(source_set.source_base)
      for source_set in project.py_libs:
        source_bases[base_path(source_set)] = linked_folder_id(source_set.source_base)

    def create_source_template(base, includes=None, excludes=None):
      return TemplateData(
        base=source_bases[base],
        includes=includes or [],
        excludes=excludes or [],
      )

    def create_sourcepath(base, sources):
      def normalize_path_pattern(path):
        return '%s/' % path if not path.endswith('/') else path

      includes = [normalize_path_pattern(src_set.path) for src_set in sources if src_set.path]
      excludes = []
      for source_set in sources:
        excludes.extend(normalize_path_pattern(exclude) for exclude in source_set.excludes)

      return create_source_template(base, includes, excludes)

    pythonpaths = []
    if project.has_python:
      for source_set in project.py_sources:
        pythonpaths.append(create_source_template(base_path(source_set)))
      for source_set in project.py_libs:
        lib_path = source_set.path if source_set.path.endswith('.egg') else '%s/' % source_set.path
        pythonpaths.append(create_source_template(base_path(source_set), includes=[lib_path]))

    configured_project = TemplateData(
      name=self.project_name,
      java=TemplateData(
        jdk=self.java_jdk,
        language_level=('1.%d' % self.java_language_level)
      ),
      has_python=project.has_python,
      has_scala=project.has_scala and not project.skip_scala,
      source_bases=source_bases.items(),
      pythonpaths=pythonpaths,
      debug_port=project.debug_port,
    )

    outdir = os.path.abspath(os.path.join(self.work_dir, 'bin'))
    safe_mkdir(outdir)

    source_sets = defaultdict(OrderedSet) # base -> source_set
    for source_set in project.sources:
      source_sets[base_path(source_set)].add(source_set)
    sourcepaths = [create_sourcepath(base, sources) for base, sources in source_sets.items()]

    libs = []
    def add_jarlibs(classpath_entries):
      for classpath_entry in classpath_entries:
        libs.append((classpath_entry.jar, classpath_entry.source_jar))
    add_jarlibs(project.internal_jars)
    add_jarlibs(project.external_jars)

    configured_classpath = TemplateData(
      sourcepaths=sourcepaths,
      has_tests=project.has_tests,
      libs=libs,
      has_scala=project.has_scala,

      # Eclipse insists the outdir be a relative path unlike other paths
      outdir=os.path.relpath(outdir, get_buildroot()),
    )

    def apply_template(output_path, template_relpath, **template_data):
      with safe_open(output_path, 'w') as output:
        Generator(pkgutil.get_data(__name__, template_relpath), **template_data).write(output)

    apply_template(self.project_filename, self.project_template, project=configured_project)
    apply_template(self.classpath_filename, self.classpath_template, classpath=configured_classpath)
    apply_template(os.path.join(self.work_dir, 'Debug on port %d.launch' % project.debug_port),
                   self.debug_template, project=configured_project)
    apply_template(self.coreprefs_filename, self.coreprefs_template, project=configured_project)

    for resource in _SETTINGS:
      with safe_open(os.path.join(self.cwd, '.settings', resource), 'w') as prefs:
        prefs.write(pkgutil.get_data(__name__, os.path.join('files', 'eclipse', resource)))

    factorypath = TemplateData(
      project_name=self.project_name,

      # The easiest way to make sure eclipse sees all annotation processors is to put all libs on
      # the apt factorypath - this does not seem to hurt eclipse performance in any noticeable way.
      jarpaths=libs
    )
    apply_template(self.apt_filename, self.apt_template, factorypath =factorypath)

    if project.has_python:
      apply_template(self.pydev_filename, self.pydev_template, project=configured_project)
    else:
      safe_delete(self.pydev_filename)

    print('\nGenerated project at %s%s' % (self.work_dir, os.sep))



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.common.collections import maybe_list

from .exclude import Exclude
from .internal import InternalTarget
from .jarable import Jarable
from .with_sources import TargetWithSources


class JvmTarget(InternalTarget, TargetWithSources, Jarable):
  """A base class for all java module targets that provides path and dependency translation."""

  def __init__(self, name, sources, dependencies,
               excludes=None,
               buildflags=None,
               configurations=None):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param sources: A list of filenames representing the source code
      this library is compiled from.
    :type sources: list of strings
    :param dependencies: List of :class:`twitter.pants.base.target.Target` instances
      this target depends on.
    :type dependencies: list of targets
    :param excludes: One or more :class:`twitter.pants.targets.exclude.Exclude` instances
      to filter this target's transitive dependencies against.
    :param buildflags: Unused, and will be removed in a future release.
    :param configurations: One or more ivy configurations to resolve for this target.
      This parameter is not intended for general use.
    :type configurations: tuple of strings
    """

    InternalTarget.__init__(self, name, dependencies)
    TargetWithSources.__init__(self, name, sources)

    self.excludes = maybe_list(excludes or [], Exclude)
    self.configurations = maybe_list(configurations or [])



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

__author__ = 'John Sirois'

import inspect
from optparse import OptionGroup

from twitter.pants.base import manual
from twitter.pants.tasks import Task


class GoalError(Exception):
  """Raised to indicate a goal has failed."""


from twitter.pants.goal.phase import Phase


class Mkflag(object):
  """A factory for namespaced flags."""

  def __init__(self, namespace):
    """Creates a new Mkflag that will use the given namespace to prefix the flags it creates.

    namespace: Either a function accepting a separator string that returns a prefix string for the
               flag or else a fixed prefix string for all flags.
    """
    self._namespace = namespace if callable(namespace) else lambda sep: namespace

  def __call__(self, name, negate=False):
    """Creates a prefixed flag with an optional negated prefix.

    name: The simple flag name to be prefixed.
    negate: True to prefix the flag with '--no-'.
    """
    return '--%s%s-%s' % ('no-' if negate else '', self._namespace('-'), name)

  def set_bool(self, option, opt_str, _, parser):
    """An Option callback to parse bool flags that recognizes the --no- negation prefix."""
    setattr(parser.values, option.dest, not opt_str.startswith("--no"))


@manual.builddict()
class Goal(object):
  def __init__(self, name, action, group=None, dependencies=None):
    self.name = name
    self.group = group
    self.dependencies = [Phase(d) for d in dependencies] if dependencies else []

    if type(action) == type and issubclass(action, Task):
      self._task = action
    else:
      args, varargs, keywords, defaults = inspect.getargspec(action)
      if varargs or keywords or defaults:
        raise GoalError('Invalid action supplied, cannot accept varargs, keywords or defaults')
      if len(args) > 2:
        raise GoalError('Invalid action supplied, must accept 0, 1, or 2 args')

      class FuncTask(Task):
        def __init__(self, context):
          Task.__init__(self, context)

          if not args:
            self.action = lambda targets: action()
          elif len(args) == 1:
            self.action = lambda targets: action(self.context)
          elif len(args) == 2:
            self.action = lambda targets: action(self.context, targets)
          else:
            raise AssertionError('Unexpected fallthrough')

        def execute(self, targets):
          self.action(targets)

      self._task = FuncTask

  @property
  def task_type(self):
    return self._task

  def setup_parser(self, phase, parser, args):
    """Allows a task to add its command line args to the global sepcification."""
    def namespace(sep):
      phase_leader = phase.goals() == [self] or self.name == phase.name
      return self.name if phase_leader else '%s%s%s' % (phase.name, sep, self.name)
    mkflag = Mkflag(namespace)

    group = OptionGroup(parser, title = namespace(':'))
    self.task_setup_parser(group, args, mkflag)
    if group.option_list:
      parser.add_option_group(group)

  def task_setup_parser(self, group, args, mkflag):
    """Allows a task to setup a parser.
    Override this method if you want to initialize the task with more goal data."""
    self._task.setup_parser(group, args, mkflag)

  def prepare(self, context):
    """Prepares a Task that can be executed to achieve this goal."""
    return self._task(context)

  def install(self, phase=None, first=False, replace=False, before=None, after=None):
    """
      Installs this goal in the specified phase (or a new phase with the same name as this Goal).

      The placement of the goal in the execution list of the phase defaults to the end but can be
      influence by specifying exactly one of the following arguments:

      first: Places this goal 1st in the phase's execution list
      replace: Replaces any existing goals in the phase with this goal
      before: Places this goal before the named goal in the phase's execution list
      after: Places this goal after the named goal in the phase's execution list
    """
    phase = Phase(phase or self.name)
    phase.install(self, first, replace, before, after)
    return phase

from twitter.pants.goal.context import Context
from twitter.pants.goal.group import Group

__all__ = (
  'Context',
  'Goal',
  'GoalError',
  'Group',
  'Phase'
)



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os
import pkgutil
import shutil
import tempfile

from xml.dom import minidom

from twitter.common.dirutil import safe_mkdir

from twitter.pants import get_buildroot, JavaTests, ScalaTests, SourceRoot
from twitter.pants.base.generator import TemplateData, Generator
from twitter.pants.tasks.ide_gen import IdeGen

__author__ = 'John Sirois'


_TEMPLATE_BASEDIR = os.path.join('templates', 'idea')


_VERSIONS = {
  '9': '12', # 9 and 12 are ipr/iml compatible
  '10': '12', # 10 and 12 are ipr/iml compatible
  '11': '12', # 11 and 12 are ipr/iml compatible
  '12': '12'
}


_SCALA_VERSION_DEFAULT = '2.9'
_SCALA_VERSIONS = {
  '2.8':                  'Scala 2.8',
  _SCALA_VERSION_DEFAULT: 'Scala 2.9',
  '2.10':                 'Scala 2.10',
  '2.10-virt':            'Scala 2.10 virtualized'
}


class IdeaGen(IdeGen):
  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    IdeGen.setup_parser(option_group, args, mkflag)

    supported_versions = sorted(list(_VERSIONS.keys()))
    option_group.add_option(mkflag("idea-version"), dest="idea_gen_version",
                            default='11', type="choice", choices=supported_versions,
                            help="[%%default] The IntelliJ IDEA version the project "
                                   "configuration should be generated for; can be one of: " \
                                   "%s" % supported_versions)

    option_group.add_option(mkflag("merge"), mkflag("merge", negate=True), default=True,
                            action="callback", callback=mkflag.set_bool, dest="idea_gen_merge",
                            help="[%default] Merge any manual customizations in existing "
                                   "Intellij IDEA configuration. If False, manual customizations "
                                   "will be over-written.")

    option_group.add_option(mkflag("open"), mkflag("open", negate=True), default=True,
                            action="callback", callback=mkflag.set_bool, dest="idea_gen_open",
                            help="[%default] Attempts top open the generated project in IDEA.")

    option_group.add_option(mkflag("bash"), mkflag("bash", negate=True), default=False,
                            action="callback", callback=mkflag.set_bool, dest="idea_gen_bash",
                            help="Adds a bash facet to the generated project configuration.")

    option_group.add_option(mkflag("scala-language-level"), default=_SCALA_VERSION_DEFAULT,
                            type="choice", choices=_SCALA_VERSIONS.keys(),
                            dest="idea_scala_language_level",
                            help="[%default] Set the scala language level used for IDEA linting.")
    option_group.add_option(mkflag("scala-maximum-heap-size"),
                            dest="idea_gen_scala_maximum_heap_size",
                            help="[%default] Sets the maximum heap size (in megabytes) for scalac.")
    option_group.add_option(mkflag("fsc"), mkflag("fsc", negate=True), default=False,
                            action="callback", callback=mkflag.set_bool, dest="idea_gen_fsc",
                            help="If the project contains any scala targets this specifies the "
                                   "fsc compiler should be enabled.")

    option_group.add_option(mkflag("java-encoding"), default="UTF-8",
                            dest="idea_gen_java_encoding",
                            help="[%default] Sets the file encoding for java files in this "
                                   "project.")
    option_group.add_option(mkflag("java-maximum-heap-size"),
                            dest="idea_gen_java_maximum_heap_size",
                            help="[%default] Sets the maximum heap size (in megabytes) for javac.")

  def __init__(self, context):
    IdeGen.__init__(self, context)


    self.intellij_output_dir = os.path.join(self.work_dir, 'out')
    self.nomerge = not context.options.idea_gen_merge
    self.open = context.options.idea_gen_open
    self.bash = context.options.idea_gen_bash

    self.scala_language_level = _SCALA_VERSIONS.get(context.options.idea_scala_language_level, None)
    self.scala_maximum_heap_size = (
      context.options.idea_gen_scala_maximum_heap_size
      or context.config.getint('idea', 'scala_maximum_heap_size_mb', default=512)
    )
    self.fsc = context.options.idea_gen_fsc

    self.java_encoding = context.options.idea_gen_java_encoding
    self.java_maximum_heap_size = (
      context.options.idea_gen_java_maximum_heap_size
      or context.config.getint('idea', 'java_maximum_heap_size_mb', default=128)
    )

    idea_version = _VERSIONS[context.options.idea_gen_version]
    self.project_template = os.path.join(_TEMPLATE_BASEDIR, 'project-%s.mk' % idea_version)
    self.module_template = os.path.join(_TEMPLATE_BASEDIR, 'module-%s.mk' % idea_version)

    self.project_filename = os.path.join(self.cwd, '%s.ipr' % self.project_name)
    self.module_filename = os.path.join(self.work_dir, '%s.iml' % self.project_name)

  def generate_project(self, project):
    def is_test_target_type(ttype):
      return issubclass(ttype, (JavaTests, ScalaTests))

    is_test_by_base = {}

    def is_test(source_set):
      if source_set.is_test:
        return True

      # Non test targets that otherwise live in test target roots (say a java_library), must
      # be marked as test for IDEA to correctly link the targets with the test code that uses
      # them.
      base = source_set.source_base
      if base not in is_test_by_base:
        is_test_by_base[base] = any(map(is_test_target_type, SourceRoot.types(base)))
      istest = is_test_by_base[base]
      if istest:
        self.context.log.debug('Marked non-test source set as test (%s, %s)' % (
          source_set.source_base, source_set.path
        ))
      return istest

    def create_content_root(source_set):
      root_relative_path = os.path.join(source_set.source_base, source_set.path) \
                           if source_set.path else source_set.source_base

      sources = TemplateData(
        path=root_relative_path,
        package_prefix=source_set.path.replace('/', '.') if source_set.path else None,
        is_test=is_test(source_set)
      )

      return TemplateData(
        path=root_relative_path,
        sources=[sources],
        exclude_paths=[os.path.join(source_set.source_base, x) for x in source_set.excludes],
      )

    content_roots = [create_content_root(source_set) for source_set in project.sources]
    if project.has_python:
      content_roots.extend(create_content_root(source_set) for source_set in project.py_sources)

    scala = None
    if project.has_scala:
      scala = TemplateData(
        language_level=self.scala_language_level,
        maximum_heap_size=self.scala_maximum_heap_size,
        fsc=self.fsc,
        compiler_classpath=project.scala_compiler_classpath
      )

    configured_module = TemplateData(
      root_dir=get_buildroot(),
      path=self.module_filename,
      content_roots=content_roots,
      has_bash=self.bash,
      has_python=project.has_python,
      has_tests=project.has_tests,
      scala=scala,
      internal_jars=[cp_entry.jar for cp_entry in project.internal_jars],
      internal_source_jars=[cp_entry.source_jar for cp_entry in project.internal_jars
                            if cp_entry.source_jar],
      external_jars=[cp_entry.jar for cp_entry in project.external_jars],
      external_source_jars=[cp_entry.source_jar for cp_entry in project.external_jars
                            if cp_entry.source_jar],
      extra_components=[],
    )

    outdir = os.path.abspath(self.intellij_output_dir)
    if not os.path.exists(outdir):
      os.makedirs(outdir)

    configured_project = TemplateData(
      root_dir=get_buildroot(),
      outdir=outdir,
      modules=[ configured_module ],
      java=TemplateData(
        encoding=self.java_encoding,
        maximum_heap_size=self.java_maximum_heap_size,
        jdk=self.java_jdk,
        language_level = 'JDK_1_%d' % self.java_language_level
      ),
      resource_extensions=project.resource_extensions,
      scala=scala,
      checkstyle_suppression_files=','.join(project.checkstyle_suppression_files),
      checkstyle_classpath=';'.join(project.checkstyle_classpath),
      debug_port=project.debug_port,
      extra_components=[],
    )

    if not self.nomerge:
      # Grab the existing components, which may include customized ones.
      existing_project_components = self._parse_xml_component_elements(self.project_filename)
      existing_module_components = self._parse_xml_component_elements(self.module_filename)

    # Generate (without merging in any extra components).
    safe_mkdir(os.path.abspath(self.intellij_output_dir))

    ipr = self._generate_to_tempfile(
        Generator(pkgutil.get_data(__name__, self.project_template), project = configured_project))
    iml = self._generate_to_tempfile(
        Generator(pkgutil.get_data(__name__, self.module_template), module = configured_module))

    if not self.nomerge:
      # Get the names of the components we generated, and then delete the
      # generated files.  Clunky, but performance is not an issue, and this
      # is an easy way to get those component names from the templates.
      extra_project_components = self._get_components_to_merge(existing_project_components, ipr)
      extra_module_components =  self._get_components_to_merge(existing_module_components, iml)
      os.remove(ipr)
      os.remove(iml)

      # Generate again, with the extra components.
      ipr = self._generate_to_tempfile(Generator(pkgutil.get_data(__name__, self.project_template),
          project = configured_project.extend(extra_components = extra_project_components)))
      iml = self._generate_to_tempfile(Generator(pkgutil.get_data(__name__, self.module_template),
          module = configured_module.extend(extra_components = extra_module_components)))

    shutil.move(ipr, self.project_filename)
    shutil.move(iml, self.module_filename)

    print('\nGenerated project at %s%s' % (self.work_dir, os.sep))

    return self.project_filename if self.open else None

  def _generate_to_tempfile(self, generator):
    """Applies the specified generator to a temp file and returns the path to that file.
    We generate into a temp file so that we don't lose any manual customizations on error."""
    (output_fd, output_path) = tempfile.mkstemp()
    with os.fdopen(output_fd, 'w') as output:
      generator.write(output)
    return output_path

  def _parse_xml_component_elements(self, path):
    """Returns a list of pairs (component_name, xml_fragment) where xml_fragment is the xml text of
    that <component> in the specified xml file."""
    if not os.path.exists(path):
      return []  # No existing components.
    dom = minidom.parse(path)
    # .ipr and .iml files both consist of <component> elements directly under a root element.
    return [ (x.getAttribute('name'), x.toxml()) for x in dom.getElementsByTagName('component') ]

  def _get_components_to_merge(self, mergable_components, path):
    """Returns a list of the <component> fragments in mergable_components that are not
    superceded by a <component> in the specified xml file.
    mergable_components is a list of (name, xml_fragment) pairs."""

    # As a convenience, we use _parse_xml_component_elements to get the
    # superceding component names, ignoring the generated xml fragments.
    # This is fine, since performance is not an issue.
    generated_component_names = set(
      [ name for (name, _) in self._parse_xml_component_elements(path) ])
    return [ x[1] for x in mergable_components if x[0] not in generated_component_names]




# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import functools
import keyword
import os
import re
import subprocess
import sys

from twitter.pants import get_buildroot
from twitter.pants.python.code_generator import CodeGenerator
from twitter.pants.python.file_copier import FileCopier
from twitter.pants.targets.python_thrift_library import PythonThriftLibrary
from twitter.pants.thrift_util import (calculate_compile_sources, find_includes,
                                       select_thrift_binary)


sys.path.append(os.path.join(get_buildroot(), 'src/python/twitter'))
from twadoop.pants.targets.remote_python_thrift_library import RemotePythonThriftLibrary


class PythonThriftBuilder(CodeGenerator):
  """Code Generator a  Python code from thrift IDL files."""
  class UnknownPlatformException(CodeGenerator.Error):
    def __init__(self, platform):
      super(PythonThriftBuilder.UnknownPlatformException, self).__init__(
          "Unknown platform: %s!" % str(platform))

  def run_thrifts(self):
    """
    Generate Python thrift code using thrift compiler specified in pants config.

    Thrift fields conflicting with Python keywords are suffixed with a trailing
    underscore (e.g.: from_).
    """
    self._workdir = os.path.join(self.config.getdefault(option='thrift_workdir'), 'py-thrift')
    def is_py_thrift(target):
      return isinstance(target, PythonThriftLibrary)
    bases, target_thrifts_sources = calculate_compile_sources([self.target], is_py_thrift)
    no_new_thrifts = False
    thrifts = target_thrifts_sources
    all_thrifts = target_thrifts_sources
    while(not no_new_thrifts):
      if not thrifts:
        no_new_thrifts = True
      else:
        inc_thrifts = set()
        for source in thrifts:
          inc_thrifts.update(find_includes(bases, source))
        thrifts = inc_thrifts.difference(thrifts)
        all_thrifts.update(inc_thrifts)

    copier = FileCopier(self._workdir)
    copied_sources = set()
    for src in all_thrifts:
      # Find and Copy the sources to work dir.
      thrift_roots = [RemotePythonThriftLibrary, PythonThriftLibrary]
      copied_sources.add(self._modify_thrift(copier.find_and_copy_relative_file(src, thrift_roots)))

    for src in copied_sources:
      if not self._run_thrift(src):
        raise PythonThriftBuilder.CodeGenerationException(
          "Could not generate .py from %s!" % src)

  def _run_thrift(self, source):
    thrift_file = source
    thrift_abs_path = os.path.abspath(thrift_file)

    args = [
      select_thrift_binary(self.config),
      '--gen',
      'py:new_style',
      '-o',
      self.codegen_root
    ]

    args.extend(['-I', self._workdir])
    args.append(thrift_abs_path)

    po = subprocess.Popen(args, cwd=self.chroot.path())
    rv = po.wait()
    if rv != 0:
      comm = po.communicate()
      print('thrift generation failed!', file=sys.stderr)
      print('STDOUT', file=sys.stderr)
      print(comm[0], file=sys.stderr)
      print('STDERR', file=sys.stderr)
      print(comm[1], file=sys.stderr)
    return rv == 0

  def _modify_thrift(self, source):
    """
    Replaces the python keywords in the thrift file

    Find all python keywords in each thrift file and appends a trailing underscore.
    For example, 'from' will be converted to 'from_'.
    """
    rewrites = []
    renames = dict((kw, '%s_' % kw) for kw in keyword.kwlist)
    token_regex = re.compile(r'(\W)(%s)(\W)' % '|'.join(renames.keys()), re.MULTILINE)
    def token_replace(match):
      return '%s%s%s' % (match.group(1), renames[match.group(2)], match.group(3))
    def replace_tokens(contents):
      return token_regex.sub(token_replace, contents)
    rewrites.append(replace_tokens)
    with open(source) as contents:
      modified = functools.reduce(lambda txt, rewrite: rewrite(txt), rewrites, contents.read())
      contents.close()
      with open(source, 'w') as thrift:
        thrift.write(modified)
    return source

  @property
  def package_dir(self):
    return "gen-py"

  def generate(self):
    # autogenerate the python files that we bundle up
    self.run_thrifts()

    # Thrift generates code with all parent namespaces with empty __init__.py's. Generally
    # speaking we want to drop anything w/o an __init__.py, and for anything with an __init__.py,
    # we want to explicitly make it a namespace package, hence the hoops here.
    for root, _, files in os.walk(os.path.normpath(self.package_root)):
      reldir = os.path.relpath(root, self.package_root)
      if reldir == '.':  # skip root
        continue
      if '__init__.py' not in files:  # skip non-packages
        continue
      init_py_abspath = os.path.join(root, '__init__.py')
      module_path = self.path_to_module(reldir)
      self.created_packages.add(module_path)
      if os.path.getsize(init_py_abspath) == 0:  # empty __init__, translate to namespace package
        with open(init_py_abspath, 'wb') as f:
          f.write(b"__import__('pkg_resources').declare_namespace(__name__)")
        self.created_namespace_packages.add(module_path)
      else:
        # non-empty __init__, this is a leaf package, usually with ttypes and constants, leave as-is
        pass

    if not self.created_packages:
      raise self.CodeGenerationException(
        'No Thrift structures declared in %s!' % self.target)



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from collections import defaultdict
from contextlib import closing
from zipfile import ZipFile

from twitter.pants.java.jar import Manifest

from .jvm_binary_task import JvmBinaryTask

from . import TaskError


class DuplicateDetector(JvmBinaryTask):
  """ Detect classes and resources with the same qualified name on the classpath. """

  @staticmethod
  def _isdir(name):
    return name[-1] == '/'

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    JvmBinaryTask.setup_parser(option_group, args, mkflag)
    option_group.add_option(mkflag("fail-fast"), mkflag("fail-fast", negate=True),
                            dest="fail_fast", default=False,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Fail fast if duplicate classes/resources are found.")

  def __init__(self, context):
    JvmBinaryTask.__init__(self, context)
    self.require_jar_dependencies()
    self.fail_fast = context.options.fail_fast

  def execute(self, targets):
    for binary in filter(self.is_binary, targets):
      self.detect_duplicates_for_target(binary)

  def detect_duplicates_for_target(self, binary_target):
    list_path = []
    for basedir, externaljar in self.list_jar_dependencies(binary_target):
      list_path.append(os.path.join(basedir, externaljar))
    self._is_conflicts(list_path, binary_target)

  def _is_conflicts(self, jar_paths, binary_target):
    artifacts_by_file_name = defaultdict(set)
    for jarpath in jar_paths:
      self.context.log.debug('  scanning %s' % jarpath)
      with closing(ZipFile(jarpath)) as zip:
        for file_name in zip.namelist():
          jar_name = os.path.basename(jarpath)
          if (not self._isdir(file_name)) and Manifest.PATH != file_name:
            artifacts_by_file_name[file_name].add(jar_name)
        zip.close()

    conflicts_by_artifacts = self._get_conflicts_by_artifacts(artifacts_by_file_name)

    if len(conflicts_by_artifacts) > 0:
      self._log_conflicts(conflicts_by_artifacts, binary_target)
      if self.fail_fast:
        raise TaskError('Failing build for target %s.' % binary_target)
      return True
    return False

  def _get_conflicts_by_artifacts(self, artifacts_by_file_name):
    conflicts_by_artifacts = defaultdict(set)
    for (file_name, artifacts) in artifacts_by_file_name.items():
      if (not artifacts) or len(artifacts) < 2: continue
      conflicts_by_artifacts[tuple(sorted(artifacts))].add(file_name)
    return conflicts_by_artifacts

  def _log_conflicts(self, conflicts_by_artifacts, target):
    self.context.log.warn('\n ===== For target %s:' % target)
    for artifacts, duplicate_files in conflicts_by_artifacts.items():
      if len(artifacts) < 2: continue
      self.context.log.warn(
        'Duplicate classes and/or resources detected in artifacts: %s' % str(artifacts))
      for duplicate_file in list(duplicate_files)[:10]:
        self.context.log.warn('     %s' % duplicate_file)




"""
The pex.pex utility builds PEX environments and .pex files specified by
sources, requirements and their dependencies.
"""

from __future__ import print_function

from optparse import OptionParser
import os
import sys

from twitter.common.python.common import safe_delete, safe_mkdtemp
from twitter.common.python.distiller import Distiller
from twitter.common.python.fetcher import Fetcher, PyPIFetcher
from twitter.common.python.installer import Installer
from twitter.common.python.interpreter import PythonInterpreter
from twitter.common.python.platforms import Platform
from twitter.common.python.resolver import resolve as requirement_resolver
from twitter.common.python.pex_builder import PEXBuilder
from twitter.common.python.pex import PEX
from twitter.common.python.tracer import Tracer


CANNOT_PARSE_REQUIREMENT = 100
CANNOT_DISTILL = 101


def die(msg, error_code=1):
  print(msg, file=sys.stderr)
  sys.exit(error_code)


def log(msg, v=False):
  if v:
    print(msg, file=sys.stderr)


def parse_bool(option, opt_str, _, parser):
  setattr(parser.values, option.dest, not opt_str.startswith('--no'))


def configure_clp():
  usage = (
      '%prog [options]\n\n'
      '%prog builds a PEX (Python Executable) file based on the given specifications: '
      'sources, requirements, their dependencies and other options')

  parser = OptionParser(usage=usage, version='%prog 0.2')

  parser.add_option(
      '--pypi', '--no-pypi',
      dest='pypi',
      default=True,
      action='callback',
      callback=parse_bool,
      help='Whether to use pypi to resolve dependencies; Default: use pypi')

  parser.add_option(
      '--python',
      dest='python',
      default=None,
      help='The Python interpreter to use to build the pex.  Either specify an explicit '
           'path to an interpreter, or specify a binary accessible on $PATH. '
           'Default: Use current interpreter.')

  parser.add_option(
      '--platform',
      dest='platform',
      default=Platform.current(),
      help='The platform for which to build the PEX.  Default: %%default')

  parser.add_option(
      '--zip-safe', '--not-zip-safe',
      dest='zip_safe',
      default=True,
      action='callback',
      callback=parse_bool,
      help='Whether or not the sources in the pex file are zip safe.  If they are '
           'not zip safe, they will be written to disk prior to execution; '
           'Default: zip safe.')

  parser.add_option(
      '--always-write-cache',
      dest='always_write_cache',
      default=False,
      action='store_true',
      help='Always write the internally cached eggs to disk prior to invoking '
           'the pex source code.  This can use less memory in RAM constrained '
           'environments. [Default: %default]')

  parser.add_option(
      '--ignore-errors',
      dest='ignore_errors',
      default=False,
      action='store_true',
      help='Ignore run-time requirement resolution errors when invoking the pex. '
           '[Default: %default]')

  parser.add_option(
      '--inherit-path',
      dest='inherit_path',
      default=False,
      action='store_true',
      help='Inherit the contents of sys.path (including site-packages) running the pex. '
           '[Default: %default]')

  parser.add_option(
      '--cache-dir',
      dest='cache_dir',
      default=os.path.expanduser('~/.pex/build'),
      help='The local cache directory to use for speeding up requirement '
           'lookups; [Default: %default]')

  parser.add_option(
      '-o', '-p', '--output-file', '--pex-name',
      dest='pex_name',
      default=None,
      help='The name of the generated .pex file: Omiting this will run PEX '
           'immediately and not save it to a file.')

  parser.add_option(
      '-e', '--entry-point',
      dest='entry_point',
      default=None,
      help='The entry point for this pex; Omiting this will enter the python '
           'REPL with sources and requirements available for import.  Can be '
           'either a module or EntryPoint (module:function) format.')

  parser.add_option(
      '-r', '--requirement',
      dest='requirements',
      metavar='REQUIREMENT',
      default=[],
      action='append',
      help='requirement to be included; may be specified multiple times.')

  parser.add_option(
      '--repo',
      dest='repos',
      metavar='PATH',
      default=[],
      action='append',
      help='Additional repository path (directory or URL) to look for requirements.')

  parser.add_option(
      '-s', '--source-dir',
      dest='source_dirs',
      metavar='DIR',
      default=[],
      action='append',
      help='Source to be packaged; This <DIR> should be a pip-installable project '
           'with a setup.py.')

  parser.add_option(
      '-v', '--verbosity',
      dest='verbosity',
      default=False,
      action='store_true',
      help='Turn on logging verbosity.')

  return parser


def build_pex(args, options):
  interpreter = None
  if options.python:
    if os.path.exists(options.python):
      interpreter = PythonInterpreter.from_binary(options.python)
    else:
      interpreter = PythonInterpreter.from_env(options.python)
    if interpreter is None:
      die('Failed to find interpreter: %s' % options.python)

  pex_builder = PEXBuilder(
      path=safe_mkdtemp(),
      interpreter=interpreter,
  )

  pex_info = pex_builder.info

  pex_info.zip_safe = options.zip_safe
  pex_info.always_write_cache = options.always_write_cache
  pex_info.ignore_errors = options.ignore_errors
  pex_info.inherit_path = options.inherit_path

  fetchers = [Fetcher(options.repos)]

  if options.pypi:
    fetchers.append(PyPIFetcher())

  resolveds = requirement_resolver(
      options.requirements,
      cache=options.cache_dir,
      fetchers=fetchers,
      interpreter=interpreter,
      platform=options.platform)

  if resolveds:
    log('Resolved distributions:', v=options.verbosity)

  for pkg in resolveds:
    log('  %s' % pkg, v=options.verbosity)
    pex_builder.add_distribution(pkg)
    pex_builder.add_requirement(pkg.as_requirement())

  for source_dir in options.source_dirs:
    dist = Installer(source_dir).distribution()
    if not dist:
      die('Failed to run installer for %s' % source_dir, CANNOT_DISTILL)
    egg_path = Distiller(dist).distill()
    if not egg_path:
      die('Failed to distill %s into egg' % dist, CANNOT_DISTILL)
    pex_builder.add_egg(egg_path)

  if options.entry_point is not None:
    log('Setting entry point to %s' % options.entry_point, v=options.verbosity)
    pex_builder.info.entry_point = options.entry_point
  else:
    log('Creating environment PEX.', v=options.verbosity)

  return pex_builder


def main():
  parser = configure_clp()
  options, args = parser.parse_args()
  verbosity = 5 if options.verbosity else -1

  with Tracer.env_override(
      PEX_VERBOSE=verbosity,
      TWITTER_COMMON_PYTHON_HTTP=verbosity,
      PYTHON_VERBOSE=verbosity):

    pex_builder = build_pex(args, options)

    if options.pex_name is not None:
      log('Saving PEX file to %s' % options.pex_name, v=options.verbosity)
      tmp_name = options.pex_name + '~'
      safe_delete(tmp_name)
      pex_builder.build(tmp_name)
      os.rename(tmp_name, options.pex_name)
      return 0

    if options.platform != Platform.current():
      log('WARNING: attempting to run PEX with differing platform!')

    pex_builder.freeze()

    log('Running PEX file at %s with args %s' % (pex_builder.path(), args), v=options.verbosity)
    pex = PEX(pex_builder.path(), interpreter=pex_builder.interpreter)
    return pex.run(args=list(args))



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import hashlib
import os

from collections import defaultdict, namedtuple

from twitter.common.collections import maybe_list, OrderedSet
from twitter.common.contextutil import open_zip
from twitter.common.dirutil import safe_open, touch
from twitter.common.lang import Compatibility

from twitter.pants import get_buildroot, is_concrete, Config
from twitter.pants.base import manual, ParseContext
from twitter.pants.targets import (
    AnonymousDeps,
    JarLibrary,
    JavaThriftLibrary,
    Pants,
    SourceRoot,
    ThriftJar,
    ThriftLibrary)
from twitter.pants.tasks import Task, TaskError


class Extract(Task):

  _EXTRACT_BASES = {}

  @classmethod
  def _extract_base(cls, jar, config=None):
    if jar not in cls._EXTRACT_BASES:
      config = config or Config.load()
      workdir = config.getdefault('pants_workdir')
      default_workdir = os.path.join(workdir, 'idl')
      extract_base = os.path.join(config.get('idl-extract', 'workdir', default=default_workdir),
                                  jar.id)
      # TODO(John Sirois): creating an empty BUILD in the source-root is a hack to to work around
      # ephemeral targets being listed and then consumed by a new pants process as inputs - fix this
      # by introducing ephemeral targets as a 1st-level concept and revert to just ensuring the
      # extract_base dir exists.
      touch(os.path.join(extract_base, 'BUILD'))
      SourceRoot.register(extract_base, JavaThriftLibrary)
      cls._EXTRACT_BASES[jar] = extract_base
    return cls._EXTRACT_BASES[jar]

  class Request(namedtuple('Request', 'target_id compiler language rpc_style namespace_sig')):
    """Represents a request to materialize a pure thrift library into a target language."""

    @classmethod
    def create(cls, target_id, compiler, language, rpc_style, namespace_map=None):
      namespace_sig = None
      if namespace_map:
        sha = hashlib.sha1()
        for ns_from, ns_to in sorted(namespace_map.items()):
          sha.update(ns_from)
          sha.update(ns_to)
        namespace_sig = sha.hexdigest()
      return cls(target_id, compiler, language, rpc_style, namespace_sig)

  # map of Request -> JavaThriftLibrary
  _PLACEHOLDER_BY_REQUEST = {}

  @classmethod
  def _register_once(cls, name, id_, compiler, language, rpc_style, namespace_map, basedir, create):
    """Create and register a ``JavaThriftLibrary`` representing a ``ThriftLibrary``
    and its thrift codegen options.

    :returns: The :class:`twitter.pants.targets.java_thrift_library.JavaThriftLibrary`
      representing the requested thrift codegen.
    """
    request = cls.Request.create(id_, compiler, language, rpc_style, namespace_map)
    if request not in cls._PLACEHOLDER_BY_REQUEST:
      with ParseContext.temp(basedir):
        unique_name = '-'.join(val for val in request._replace(target_id=name) if val is not None)
        placeholder = create(unique_name)
        assert(isinstance(placeholder, JavaThriftLibrary))
      cls._PLACEHOLDER_BY_REQUEST[request] = placeholder
    return cls._PLACEHOLDER_BY_REQUEST[request]

  _PLACEHOLDERS_BY_JAR = defaultdict(list)
  _REGISTERED_JARS = set()

  @classmethod
  def _register_jar(cls, jar, compiler=None, language=None, rpc_style=None, generated_deps=None,
                    namespace_map=None, config=None):

    def create(name):
      synthetic = JavaThriftLibrary(name,
                                    sources=None,
                                    dependencies=[jar] + (generated_deps or []),
                                    compiler=compiler,
                                    language=language,
                                    rpc_style=rpc_style,
                                    namespace_map=namespace_map)
      synthetic.derived_from = jar
      cls._PLACEHOLDERS_BY_JAR[jar].append(synthetic)
      cls._REGISTERED_JARS.add((jar.org, jar.name))
      return synthetic
    return cls._register_once(jar.id, jar.id, compiler, language, rpc_style, namespace_map,
                              cls._extract_base(jar, config=config), create)

  @classmethod
  def is_registered_jar(cls, org, name):
    """Check if the jar is registered.

    :param string org: Maven groupId of the jar.
    :param string name: Maven artifactId of the jar.
    :returns: True if the jar is registered.
    :rtype: bool
    """
    assert(isinstance(org, Compatibility.string))
    assert(isinstance(name, Compatibility.string))
    return (org, name) in cls._REGISTERED_JARS

  @classmethod
  def _register_library(cls, library, compiler=None, language=None, rpc_style=None,
                        generated_deps=None, namespace_map=None):
    """Register a JavaThriftLibrary corresponding to the given ThriftLibrary.

    :param library: The :class:`twitter.pants.targets.thrift_library.ThriftLibrary`
      to register.
    :param string compiler: Name of the thrift compiler to generate sources with.
    :param string language: Language to generate code for
      (only supported by some compilers).
    :param string rpc_style: Style of RPC service stubs to generate
      (only supported by some compilers).
    :param generated_deps: Dependencies for the code that will be generated
      from the ``idl_deps``.
    :param namespace_map: A mapping from IDL declared namespaces to custom
      namespaces (only supported by some compilers).
    :returns: A :class:`twitter.pants.targets.java_thrift_library.JavaThriftLibrary`
      target that owns the generated code.
    """

    assert(isinstance(library, ThriftLibrary))
    basedir = library.address.buildfile.parent_path

    def create(name):
      # TODO(John Sirois): make the internals of targets and source path ownership easier to handle
      # with utilities or mixins that consolidate this handling centrally.

      # Sources should be relative to the BUILD file, so we need to do some math since paths go in
      # relative and come out (library.sources) qualified.
      #
      # In particular, given:
      #  source_root: src/thrift
      #  BUILD: com/twitter/base/BUILD
      #  thrift BUILD relative path: v1/api.thrift
      #
      # Such that we have the source file: src/thrift/com/twitter/base/v1/api.thrift
      # And the target: src/thrift/com/twitter/base:api-v1
      #
      # The BUILD target has input sources:
      #  lib = thrift_library(name='api-v1', sources=['v1/api.thrift'])
      #
      # Which is resolved such that:
      #   assert lib.sources == ['com/twitter/base/v1/api.thrift']
      #
      # IE: the target sources are relative to the target_base
      sources = [os.path.relpath(os.path.join(get_buildroot(), library.target_base, path), basedir)
                 for path in library.sources]

      synthetic = JavaThriftLibrary('__compiled_idl_%s' % (name),
                                    sources=sources,
                                    dependencies=generated_deps,
                                    compiler=compiler,
                                    language=language,
                                    rpc_style=rpc_style,
                                    namespace_map=namespace_map)
      synthetic.derived_from = library
      return synthetic
    return cls._register_once(library.name, library.id, compiler, language, rpc_style,
                              namespace_map, basedir, create)

  @classmethod
  @manual.builddict()
  def compiled_idl(cls, idl_deps, name=None, compiler=None, language=None, rpc_style=None,
                   generated_deps=None, namespace_map=None):
    """Now-deprecated way to generate code from Thrift IDL.
    You probably want ``java_thrift_library`` instead.

    ``compiled_idl`` was used in two main ways:

    * Creating an addressable target that may be depended on by other targets
      (this requires specifying the ``name`` parameter).
    * Creating an inline anonymous target. This was typically used in
      conjunction with the dangerous ``namespace_map`` parameter.

    :param idl_deps: One or more dependencies resolvable to a set of
      :class:`twitter.pants.targets.thrift_library.ThriftJar` or
      :class:`twitter.pants.targets.thrift_library.ThriftLibrary` targets.
    :param string name: Name of the returned target so it can be
      referenced. Anonymous (un-addressable) by default.
    :param string compiler: Name of the thrift compiler to generate sources with.
    :param string language: Language to generate code for
      (only supported by some compilers).
    :param string rpc_style: Style of RPC service stubs to generate
      (only supported by some compilers).
    :param generated_deps: Dependencies for the code that will be generated
      from the ``idl_deps``.
    :param namespace_map: Dangerous mapping from IDL declared namespaces to
      custom namespaces. This feature only supported by some code generators.
      Why "dangerous"? Folks who non-Pants-ly generate code from your Thrift
      probably don't know about this namespace map, and are likely to use
      the namespace built into the Thrift IDL instead; you should set up a
      namespace in the Thrift IDL instead.
    :returns: A :class:`twitter.pants.targets.jar_library.JarLibrary` that
      depends on
      :class:`twitter.pants.targets.java_thrift_library.JavaThriftLibrary`
      targets that own the generated code.
    """
    def stitch_dependencies(synthetic_jarlib):
      thrift_jars = OrderedSet()
      thrift_libraries = OrderedSet()
      invalid = OrderedSet()

      def is_thrift_jar(dep):
        return isinstance(dep, ThriftJar)

      def accumulate(dep):
        if is_thrift_jar(dep):
          thrift_jars.add(dep)
        elif isinstance(dep, ThriftLibrary):
          thrift_libraries.add(dep)
          thrift_jars.update(filter(is_thrift_jar, dep.dependencies))
        else:
          invalid.add(dep)

      for idl_dep in maybe_list(idl_deps,
                                expected_type=(JarLibrary, Pants, ThriftJar, ThriftLibrary),
                                raise_type=TaskError):
        idl_dep.walk(accumulate, predicate=is_concrete)

      if invalid:
        raise TaskError('Can only arrange for compiled idl from thrift jars and thrift libraries, '
                        'found the following non-compliant dependencies:\n\t%s'
                        % '\n\t'.join(map(str, invalid)))

      synthetic_by_idl = {}
      for thrift_jar in thrift_jars:
        synthetic_by_idl[thrift_jar] = cls._register_jar(thrift_jar, compiler, language, rpc_style,
                                                         generated_deps, namespace_map)

      for thrift_library in thrift_libraries:
        synthetic_by_idl[thrift_library] = cls._register_library(thrift_library, compiler, language,
                                                                 rpc_style, generated_deps,
                                                                 namespace_map)

      # Stitch dependencies between the synthetic targets that mirror the IDL inter-dependencies.
      for thrift_library in thrift_libraries:
        synthetic_deps = OrderedSet()
        for dependency in thrift_library.dependencies:
          for dep in dependency.resolve():
            if is_concrete(dep):
              synthetic_deps.add(synthetic_by_idl[dep])
        synthetic_by_idl[thrift_library].update_dependencies(synthetic_deps)
      synthetic_jarlib.dependencies.update(synthetic_by_idl.values())

    synthetic_lib = JarLibrary(name=name, dependencies=()) if name else AnonymousDeps()
    ParseContext.locate().on_context_exit(stitch_dependencies, synthetic_lib)
    return synthetic_lib

  def __init__(self, context):
    Task.__init__(self, context)

    self._placeholders = {}
    compiled_idl = set()
    for thrift_jar, placeholders in self._PLACEHOLDERS_BY_JAR.items():
      # When a BUILD file is eval'ed all its target constructors get run potentially registering
      # compiled_idl jars that are not actually used in the active target graph - guard for this
      # case and only act on those placeholders active in the context.
      active_placeholders = self.context.targets(predicate=placeholders.__contains__)
      if active_placeholders:
        # Any representative active placeholder for the jar will do for resolving, the compiler,
        # language, etc. do not come into play here; so we pick just one to minimize resolves.
        representative = active_placeholders[0]
        self.context.log.debug('Fetching thrift jar %s via representative %s'
                               % (thrift_jar, representative))
        compiled_idl.add(representative)

        self._placeholders[(thrift_jar, representative)] = active_placeholders

    def is_compiled_idl(target):
      return target in compiled_idl

    context.products.require('idl_dependencies', predicate=is_compiled_idl)

  def execute(self, _):
    depmap = self.context.products.get('idl_dependencies')
    for (thrift_jar, representative), placeholders in self._placeholders.items():
      sources = set()
      extract_base = self._extract_base(thrift_jar, config=self.context.config)
      for basedir, jars in depmap.get(representative).items():
        for jar in jars:
          sources.update(self._extract(extract_base, os.path.join(basedir, jar)))

      for placeholder in placeholders:
        placeholder.sources = sources

  def _extract(self, extract_base, jar_path):
    self.context.log.debug('Extracting idl jar %s to: %s' % (jar_path, extract_base))

    with open_zip(jar_path) as jar:
      sources = set()
      for path in jar.namelist():
        if path.endswith('.thrift'):
          contents = jar.read(path)
          with safe_open(os.path.join(extract_base, path), 'w') as out:
            out.write(contents)
          sources.add(path)

      self.context.log.debug('Found thrift IDL sources: %s' % sources)
      return sources



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import glob
import os
import subprocess
import sys

from twitter.common.contextutil import pushd
from twitter.common.python.installer import Packager


class SdistBuilder(object):
  """A helper class to run setup.py projects."""

  class Error(Exception): pass
  class SetupError(Error): pass

  @classmethod
  def build(cls, setup_root, target, interpreter=None):
    packager = Packager(setup_root, interpreter=interpreter,
        install_dir=os.path.join(setup_root, 'dist'))
    try:
      return packager.sdist()
    except Packager.InstallFailure as e:
      raise cls.SetupError(str(e))



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from twitter.common.collections import OrderedSet

from twitter.pants import is_scala, is_test
from twitter.pants.binary_util import safe_args
from twitter.pants.ivy import profile_classpath
from twitter.pants.java import runjava

from .jvm_task import JvmTask

from . import Task, TaskError


class SpecsRun(JvmTask):
  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    option_group.add_option(mkflag("skip"), mkflag("skip", negate=True), dest = "specs_run_skip",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%default] Skip running specs")

    option_group.add_option(mkflag("debug"), mkflag("debug", negate=True), dest = "specs_run_debug",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%default] Run specs with a debugger")

    option_group.add_option(mkflag("jvmargs"), dest = "specs_run_jvmargs", action="append",
                            help = "Runs specs in a jvm with these extra jvm args.")

    option_group.add_option(mkflag("test"), dest = "specs_run_tests", action="append",
                            help = "[%default] Force running of just these specs.  Tests can be "
                                   "specified either by fully qualified classname or "
                                   "full file path.")

    option_group.add_option(mkflag("color"), mkflag("color", negate=True),
                            dest="specs_run_color", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Emit test result with ANSI terminal color codes.")

  def __init__(self, context):
    Task.__init__(self, context)

    self.profile = context.config.get('specs-run', 'profile')
    self.confs = context.config.getlist('specs-run', 'confs')

    self.jvm_args = context.config.getlist('specs-run', 'jvm_args', default=[])
    if context.options.specs_run_jvmargs:
      self.jvm_args.extend(context.options.specs_run_jvmargs)
    if context.options.specs_run_debug:
      self.jvm_args.extend(context.config.getlist('jvm', 'debug_args'))

    self.skip = context.options.specs_run_skip
    self.color = context.options.specs_run_color

    self.workdir = context.config.get('specs-run', 'workdir')

    self.tests = context.options.specs_run_tests

  def execute(self, targets):
    if not self.skip:
      def run_tests(tests):
        args = ['--color'] if self.color else []
        args.append('--specs=%s' % ','.join(tests))
        specs_runner_main = 'com.twitter.common.testing.ExplicitSpecsRunnerMain'

        _, classpath = profile_classpath(self.profile)
        result = runjava(
          classpath=self.classpath(classpath, confs=self.confs),
          main=specs_runner_main,
          args=args,
          jvm_args=self.jvm_args
        )
        if result != 0:
          raise TaskError('java %s ... exited non-zero (%i)' % (specs_runner_main, result))

      if self.tests:
        run_tests(self.tests)
      else:
        with safe_args(self.calculate_tests(targets)) as tests:
          if tests:
            run_tests(tests)

  def calculate_tests(self, targets):
    tests = OrderedSet()
    for target in targets:
      if is_scala(target) and is_test(target):
        tests.update(os.path.join(target.target_base, test) for test in target.sources)
    return tests



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os
import sys

from abc import abstractmethod
from collections import defaultdict

from twitter.common.lang import AbstractClass, Compatibility

from twitter.pants import get_buildroot, get_scm, has_resources, has_sources, TaskError
from twitter.pants.base.build_file import BuildFile
from twitter.pants.base.target import Target
from twitter.pants.scm import Scm
from twitter.pants.tasks.console_task import ConsoleTask


class WhatChanged(ConsoleTask):
  """Emits the targets that have been modified since a given commit."""

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    super(WhatChanged, cls).setup_parser(option_group, args, mkflag)

    option_group.add_option(mkflag('parent'), dest='what_changed_create_prefix', default='HEAD',
                            help='[%default] Identifies the parent tree-ish to calculate changes '
                                 'against.')

    option_group.add_option(mkflag("files"), mkflag("files", negate=True), default=False,
                            action="callback", callback=mkflag.set_bool,
                            dest='what_changed_show_files',
                            help='[%default] Shows changed files instead of the targets that own '
                                 'them.')

  def __init__(self, context, workspace, outstream=sys.stdout):
    if not isinstance(workspace, Workspace):
      raise ValueError('WhatChanged requires a Workspace, given %s' % workspace)

    super(WhatChanged, self).__init__(context, outstream)

    self._workspace = workspace

    self._parent = context.options.what_changed_create_prefix
    self._show_files = context.options.what_changed_show_files

    self._filemap = defaultdict(set)

  def console_output(self, _):
    touched_files = self._get_touched_files()
    if self._show_files:
      for file_ in touched_files:
        yield file_
    else:
      touched_targets = set()
      for file_ in touched_files:
        for touched_target in self._owning_targets(file_):
          if touched_target not in touched_targets:
            touched_targets.add(touched_target)
            yield str(touched_target.address)

  def _get_touched_files(self):
    try:
      return self._workspace.touched_files(self._parent)
    except Workspace.WorkspaceError as e:
      raise TaskError(e)

  def _owning_targets(self, file_):
    for build_file in self._candidate_owners(file_):
      is_build_file = (build_file.full_path == os.path.join(get_buildroot(), file_))
      for address in Target.get_all_addresses(build_file):
        target = Target.get(address)

        # A synthesized target can never own permanent files on disk
        if target != target.derived_from:
          # TODO(John Sirois): tighten up the notion of targets written down in a BUILD by a user
          # vs. targets created by pants at runtime.
          continue

        if target and (is_build_file or ((has_sources(target) or has_resources(target))
                                         and self._owns(target, file_))):
          yield target

  def _candidate_owners(self, file_):
    build_file = BuildFile(get_buildroot(), relpath=os.path.dirname(file_), must_exist=False)
    if build_file.exists():
      yield build_file
    for sibling in build_file.siblings():
      yield sibling
    for ancestor in build_file.ancestors():
      yield ancestor

  def _owns(self, target, file_):
    if target not in self._filemap:
      files = self._filemap[target]
      files_owned_by_target = target.sources if has_sources(target) else []
      # TODO (tdesai): This case to handle resources in PythonTarget.
      # Remove this when we normalize resources handling across python and jvm targets.
      if has_resources(target):
        for resource in target.resources:
          if isinstance(resource, Compatibility.string):
            files_owned_by_target.extend(target.resources)
      for owned_file in files_owned_by_target:
        owned_path = os.path.join(target.target_base, owned_file)
        files.add(owned_path)
    return file_ in self._filemap[target]


class Workspace(AbstractClass):
  """Tracks the state of the current workspace."""

  class WorkspaceError(Exception):
    """Indicates a problem reading the local workspace."""

  @abstractmethod
  def touched_files(self, parent):
    """Returns the set of paths modified between the given parent commit and the current local
    workspace state.
    """


class ScmWorkspace(Workspace):
  """A workspace that uses an Scm to determine the touched files."""

  def __init__(self, scm):
    super(ScmWorkspace, self).__init__()

    self._scm = scm or get_scm()

  def touched_files(self, parent):
    try:
      return self._scm.changed_files(from_commit=parent, include_untracked=True)
    except Scm.ScmException as e:
      raise self.WorkspaceError("Problem detecting changed files.", e)


class ScmWhatChanged(WhatChanged):
  def __init__(self, context, scm=None, outstream=sys.stdout):
    """Creates a WhatChanged task that uses an Scm to determine changed files.

    context:    The pants execution context.
    scm:        The scm to use, taken from the globally configured scm if None.
    outstream:  The stream to write changed files or targets to.
    """
    super(ScmWhatChanged, self).__init__(context, ScmWorkspace(scm or get_scm()), outstream)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

__all__ = (
    'Address',
    'BuildFile',
    'Config',
    'get_builddict_info',
    'manual',
    'ParseContext',
    'Target',
    'TargetDefinitionException'
)

from .address import Address
from .build_manual import get_builddict_info, manual
from .build_file import BuildFile
from .config import Config
from .parse_context import ParseContext
from .target import Target, TargetDefinitionException



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.pants.targets import IdlJvmThriftLibrary, JavaThriftLibrary
from twitter.pants.tasks import Task


class IdlProvidedBy(Task):
  """ Replaces idl jar dependency with correct published jar

    Replaces this target with the jar specified via the constructor arg 'provided_by'
    in the dependency graph in the given context.  This is necessary if you are attempting
    to publish an artifact that depends on an IdlJvmThriftLibrary target, which doesn't
    provide an artifact directly.
  """

  def __init__(self, context):
    Task.__init__(self, context)
    self.context.products.require('java')

  def _cleanup_target(self, target, compiled_jar, idl_jar):
    for dependant in self.context.dependants(on_predicate=lambda tgt: tgt == target):
      if hasattr(dependant, 'replace_dependency'):
        if isinstance(dependant, JavaThriftLibrary):
          self.context.log.info(
            'Replacing %s with %s as dependency of %s' % (target, idl_jar, dependant))
          idl_jar.idl_only = True
          dependant.replace_dependency(target, idl_jar)
        else:
          self.context.log.info(
            'Replacing %s with %s as dependency of %s' % (target, compiled_jar, dependant))
          dependant.replace_dependency(target, compiled_jar)
    self.context.remove_target(target)

  def execute(self, targets):
    def process(target):
      compiled_jar = target.provided_by
      idl_jar = target.idl_jar
      self._cleanup_target(target, compiled_jar, idl_jar)
      genmap = self.context.products.get('java').get(target)
      if genmap:
        for _, java_targets in genmap.items():
          for java_target in java_targets:
            self._cleanup_target(java_target, compiled_jar, idl_jar)

    for target in targets:
      target.walk(process, predicate=lambda target: isinstance(target, IdlJvmThriftLibrary))



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from twitter.common.dirutil import safe_mkdir
from twitter.common.python.distiller import Distiller
from twitter.common.python.http.link import SourceLink
from twitter.common.python.installer import Installer
from twitter.common.python.interpreter import PythonInterpreter, PythonIdentity
from twitter.common.python.obtainer import Obtainer

from .resolver import crawler_from_config, fetchers_from_config

from pkg_resources import Requirement


# TODO(wickman) Create a safer version of this and add to twitter.common.dirutil
def safe_link(src, dst):
  try:
    os.unlink(dst)
  except OSError:
    pass
  os.symlink(src, dst)


class PythonInterpreterCache(object):
  def __init__(self, config, logger=None):
    self._path = config.get('python-setup', 'interpreter_cache')
    setuptools_req = 'setuptools==%s' % config.get('python-setup', 'setuptools_version')
    try:
      self._setuptools_requirement = Requirement.parse(setuptools_req, replacement=False)
    except TypeError:
      self._setuptools_requirement = Requirement.parse(setuptools_req)
    safe_mkdir(self._path)
    self._fetchers = fetchers_from_config(config)
    self._crawler = crawler_from_config(config)
    self._interpreters = set()
    self._logger = logger or (lambda msg: True)

  @property
  def interpreters(self):
    return self._interpreters

  @classmethod
  def interpreter_from_path(cls, path):
    interpreter_dir = os.path.basename(path)
    identity = PythonIdentity.from_path(interpreter_dir)
    try:
      executable = os.readlink(os.path.join(path, 'python'))
    except OSError:
      return None
    try:
      distribute_path = os.readlink(os.path.join(path, 'distribute'))
    except OSError:
      distribute_path = None
    return PythonInterpreter(executable, identity, distribute_path)

  def setup_distribute(self, interpreter, dest):
    obtainer = Obtainer(self._crawler, self._fetchers, [])
    obtainer_iterator = obtainer.iter(self._setuptools_requirement)
    links = [link for link in obtainer_iterator if isinstance(link, SourceLink)]
    for link in links:
      self._logger('Fetching %s' % link)
      sdist = link.fetch()
      self._logger('Installing %s' % sdist)
      installer = Installer(sdist, strict=False, interpreter=interpreter)
      dist = installer.distribution()
      self._logger('Distilling %s' % dist)
      egg = Distiller(dist).distill(into=dest)
      safe_link(egg, os.path.join(dest, 'distribute'))
      break

  def setup_interpreter(self, interpreter):
    interpreter_dir = os.path.join(self._path, str(interpreter.identity))
    safe_mkdir(interpreter_dir)
    safe_link(interpreter.binary, os.path.join(interpreter_dir, 'python'))
    if interpreter.distribute:
      safe_link(interpreter.distribute, os.path.join(interpreter_dir, 'distribute'))
    else:
      self.setup_distribute(interpreter, interpreter_dir)

  def setup_cached(self):
    for interpreter_dir in os.listdir(self._path):
      path = os.path.join(self._path, interpreter_dir)
      pi = self.interpreter_from_path(path)
      if pi:
        self._logger('Found interpreter %s: %s (cached)' % (pi.binary, str(pi.identity)))
        self._interpreters.add(pi)

  def setup_paths(self, paths):
    for interpreter in PythonInterpreter.all(paths):
      identity_str = str(interpreter.identity)
      path = os.path.join(self._path, identity_str)
      pi = self.interpreter_from_path(path)
      if pi is None or pi.distribute is None:
        self._logger('Found interpreter %s: %s (%s)' % (
            interpreter.binary,
            str(interpreter.identity),
            'uncached' if pi is None else 'incomplete'))
        self.setup_interpreter(interpreter)
        pi = self.interpreter_from_path(path)
        if pi is None or pi.distribute is None:
          continue
      self._interpreters.add(pi)

  def matches(self, filters):
    for interpreter in self._interpreters:
      if any(interpreter.identity.matches(filt) for filt in filters):
        yield interpreter

  def setup(self, paths=(), force=False, filters=('',)):
    has_setup = False
    setup_paths = paths or os.getenv('PATH').split(os.pathsep)
    self.setup_cached()
    if force:
      has_setup = True
      self.setup_paths(setup_paths)
    matches = list(self.matches(filters))
    if len(matches) == 0 and not has_setup:
      self.setup_paths(setup_paths)
      matches = list(self.matches(filters))
    if len(matches) == 0:
      self._logger('Found no valid interpreters!')
    return matches

  def select_interpreter(self, compatibilities, allow_multiple=False):
    if allow_multiple:
      return compatibilities
    me = PythonInterpreter.get()
    if me in compatibilities:
      return [me]
    return [min(compatibilities)] if compatibilities else []



from __future__ import print_function

import os
import time

from twitter.common.dirutil import touch
from twitter.common.python.fetcher import Fetcher, PyPIFetcher
from twitter.common.python.http import Crawler
from twitter.common.python.obtainer import Obtainer
from twitter.common.python.interpreter import PythonInterpreter
from twitter.common.python.platforms import Platform
from twitter.common.python.resolver import requirement_is_exact
from twitter.common.python.translator import (
    ChainedTranslator,
    EggTranslator,
    SourceTranslator)

from pkg_resources import (
    Environment,
    WorkingSet,
)


def get_platforms(platform_list):
  def translate(platform):
    return Platform.current() if platform == 'current' else platform
  return tuple(set(map(translate, platform_list)))


def fetchers_from_config(config):
  fetchers = []
  fetchers.extend(Fetcher([url]) for url in config.getlist('python-repos', 'repos', []))
  fetchers.extend(PyPIFetcher(url) for url in config.getlist('python-repos', 'indices', []))
  return fetchers


def crawler_from_config(config, conn_timeout=None):
  return Crawler(cache=config.get('python-setup', 'download_cache'), conn_timeout=conn_timeout)


class PantsEnvironment(Environment):

  # TODO(wickman) Use the twitter.common.python version once wheels are implemented there.
  def can_add(self, dist):
    return Platform.distribution_compatible(dist, python=self.python, platform=self.platform)


def resolve_multi(config,
                  requirements,
                  interpreter=None,
                  platforms=None,
                  conn_timeout=None,
                  ttl=3600):
  """Multi-platform dependency resolution for PEX files.

     Given a pants configuration and a set of requirements, return a list of distributions
     that must be included in order to satisfy them.  That may involve distributions for
     multiple platforms.

     :param config: Pants :class:`Config` object.
     :param requirements: A list of :class:`PythonRequirement` objects to resolve.
     :param interpreter: :class:`PythonInterpreter` for which requirements should be resolved.
                         If None specified, defaults to current interpreter.
     :param platforms: Optional list of platforms against requirements will be resolved. If
                         None specified, the defaults from `config` will be used.
     :param conn_timeout: Optional connection timeout for any remote fetching.
     :param ttl: Time in seconds before we consider re-resolving an open-ended requirement, e.g.
                 "flask>=0.2" if a matching distribution is available on disk.  Defaults
                 to 3600.
  """
  now = time.time()
  distributions = {}

  interpreter = interpreter or PythonInterpreter.get()
  if not isinstance(interpreter, PythonInterpreter):
    raise TypeError('Expected interpreter to be a PythonInterpreter, got %s' % type(interpreter))

  install_cache = config.get('python-setup', 'install_cache')
  platforms = get_platforms(platforms or config.getlist('python-setup', 'platforms', ['current']))
  crawler = crawler_from_config(config, conn_timeout=conn_timeout)
  fetchers = fetchers_from_config(config)

  for platform in platforms:
    env = PantsEnvironment(search_path=[], platform=platform, python=interpreter.python)
    working_set = WorkingSet(entries=[])

    shared_options = dict(install_cache=install_cache, platform=platform)
    egg_translator = EggTranslator(python=interpreter.python, **shared_options)
    egg_obtainer = Obtainer(crawler, [Fetcher([install_cache])], egg_translator)

    def installer(req):
      # Attempt to obtain the egg from the local cache.  If it's an exact match, we can use it.
      # If it's not an exact match, then if it's been resolved sufficiently recently, we still
      # use it.
      dist = egg_obtainer.obtain(req)
      if dist and (requirement_is_exact(req) or now - os.path.getmtime(dist.location) < ttl):
        return dist

      # Failed, so follow through to "remote" resolution
      source_translator = SourceTranslator(
           interpreter=interpreter,
           use_2to3=getattr(req, 'use_2to3', False),
           **shared_options)
      translator = ChainedTranslator(egg_translator, source_translator)
      obtainer = Obtainer(
          crawler,
          [Fetcher([req.repository])] if getattr(req, 'repository', None) else fetchers,
          translator)
      dist = obtainer.obtain(req)
      if dist:
        try:
          touch(dist.location)
        except OSError:
          pass
      return dist

    distributions[platform] = working_set.resolve(requirements, env=env, installer=installer)

  return distributions



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

__author__ = 'John Sirois'

import os

from twitter.pants import get_buildroot, is_jvm, is_test
from twitter.pants.tasks import Task


class JvmTask(Task):
  def classpath(self, cp=None, confs=None):
    classpath = cp or []
    with self.context.state('classpath', []) as cp:
      classpath.extend(path for conf, path in cp if not confs or conf in confs)

    if self.context.config.getbool('jvm', 'parallel_test_paths', default=False):
      # TODO(John Sirois): Kill this when science tests are converted to use explicit test target
      # resources: http://jira.local.twitter.com/browse/AWESOME-108
      bases = set()
      for target in self.context.targets():
        if is_jvm(target) and is_test(target):
          if target.target_base not in bases:
            sibling_resources_base = os.path.join(os.path.dirname(target.target_base), 'resources')
            classpath.append(os.path.join(get_buildroot(), sibling_resources_base))
            bases.add(target.target_base)

    return classpath



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import errno
import os
import re
import subprocess

from collections import defaultdict, namedtuple

from twitter.common import log
from twitter.common.collections import OrderedSet
from twitter.common.dirutil import safe_mkdir

from twitter.pants import get_buildroot, is_jvm, is_python
from twitter.pants.targets import (
  InternalTarget,
  JavaLibrary,
  JavaThriftLibrary,
  PythonLibrary,
  PythonThriftLibrary)
from twitter.pants.tasks import TaskError
from twitter.pants.tasks.code_gen import CodeGen
from twitter.pants.thrift_util import calculate_compile_roots, select_thrift_binary


def _copytree(from_base, to_base):
  def abort(error):
    raise TaskError('Failed to copy from %s to %s: %s' % (from_base, to_base, error))

  # TODO(John Sirois): Consider adding a unit test and lifting this to common/dirutils or similar
  def safe_link(src, dst):
    try:
      os.link(src, dst)
    except OSError as e:
      if e.errno != errno.EEXIST:
        raise e

  for dirpath, dirnames, filenames in os.walk(from_base, topdown=True, onerror=abort):
    to_path = os.path.join(to_base, os.path.relpath(dirpath, from_base))
    for dirname in dirnames:
      safe_mkdir(os.path.join(to_path, dirname))
    for filename in filenames:
      safe_link(os.path.join(dirpath, filename), os.path.join(to_path, filename))


class ThriftGen(CodeGen):
  GenInfo = namedtuple('GenInfo', ['gen', 'deps'])
  ThriftSession = namedtuple('ThriftSession', ['outdir', 'cmd', 'process'])

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    option_group.add_option(mkflag("outdir"), dest="thrift_gen_create_outdir",
                            help="Emit generated code in to this directory.")

    option_group.add_option(mkflag("version"), dest="thrift_version",
                            help="Thrift compiler version.")

    option_group.add_option(mkflag("lang"), dest="thrift_gen_langs",  default=[],
                            action="append", type="choice", choices=['python', 'java'],
                            help="Force generation of thrift code for these languages.  Both "
                                 "'python' and 'java' are supported")

  def __init__(self, context):
    CodeGen.__init__(self, context)

    output_dir = (
      context.options.thrift_gen_create_outdir
      or context.config.get('thrift-gen', 'workdir')
    )
    self.combined_dir = os.path.join(output_dir, 'combined')
    self.session_dir = os.path.join(output_dir, 'sessions')

    self.strict = context.config.getbool('thrift-gen', 'strict')
    self.verbose = context.config.getbool('thrift-gen', 'verbose')

    def create_geninfo(key):
      gen_info = context.config.getdict('thrift-gen', key)
      gen = gen_info['gen']
      deps = {}
      for category, depspecs in gen_info['deps'].items():
        dependencies = OrderedSet()
        deps[category] = dependencies
        for depspec in depspecs:
          dependencies.update(context.resolve(depspec))
      return self.GenInfo(gen, deps)

    self.gen_java = create_geninfo('java')
    self.gen_python = create_geninfo('python')

    self.gen_langs = set(context.options.thrift_gen_langs)
    for lang in ('java', 'python'):
      if self.context.products.isrequired(lang):
        self.gen_langs.add(lang)


  def invalidate_for(self):
    return self.gen_langs

  def is_gentarget(self, target):
    return ((isinstance(target, JavaThriftLibrary) and target.compiler == 'thrift')
            or isinstance(target, PythonThriftLibrary))

  def is_forced(self, lang):
    return lang in self.gen_langs

  def genlangs(self):
    return dict(java=is_jvm, python=is_python)

  def genlang(self, lang, targets):
    thrift_binary = select_thrift_binary(
      self.context.config,
      version=self.context.options.thrift_version
    )

    bases, sources = calculate_compile_roots(targets, self.is_gentarget)

    if lang == 'java':
      gen = self.gen_java.gen
    elif lang == 'python':
      gen = self.gen_python.gen
    else:
      raise TaskError('Unrecognized thrift gen lang: %s' % lang)

    args = [
      thrift_binary,
      '--gen', gen,
      '-recurse',
    ]

    if self.strict:
      args.append('-strict')
    if self.verbose:
      args.append('-verbose')
    for base in bases:
      args.extend(('-I', base))

    sessions = []
    for source in sources:
      # Create a unique session dir for this thrift root.  Sources may be full paths but we only
      # need the path relative to the build root to ensure uniqueness.
      # TODO(John Sirois): file paths should be normalized early on and uniformly, fix the need to
      # relpath here at all.
      relsource = os.path.relpath(source, get_buildroot())
      outdir = os.path.join(self.session_dir, '.'.join(relsource.split(os.path.sep)))
      safe_mkdir(outdir)

      cmd = args[:]
      cmd.extend(('-o', outdir))
      cmd.append(source)
      log.debug('Executing: %s' % ' '.join(cmd))
      sessions.append(self.ThriftSession(outdir, cmd, subprocess.Popen(cmd)))

    result = 0
    for session in sessions:
      if result != 0:
        session.process.kill()
      else:
        result = session.process.wait()
        if result != 0:
          self.context.log.error('Failed: %s' % ' '.join(session.cmd))
        else:
          _copytree(session.outdir, self.combined_dir)
    if result != 0:
      raise TaskError('%s ... exited non-zero (%i)' % (thrift_binary, result))

  def createtarget(self, lang, gentarget, dependees):
    if lang == 'java':
      return self._create_java_target(gentarget, dependees)
    elif lang == 'python':
      return self._create_python_target(gentarget, dependees)
    else:
      raise TaskError('Unrecognized thrift gen lang: %s' % lang)

  def _create_java_target(self, target, dependees):
    def create_target(files, deps):
       return self.context.add_new_target(os.path.join(self.combined_dir, 'gen-java'),
                                          JavaLibrary,
                                          name=target.id,
                                          provides=target.provides,
                                          sources=files,
                                          dependencies=deps)
    return self._inject_target(target, dependees, self.gen_java, 'java', create_target)

  def _create_python_target(self, target, dependees):
    def create_target(files, deps):
     return self.context.add_new_target(os.path.join(self.combined_dir, 'gen-py'),
                                        PythonLibrary,
                                        name=target.id,
                                        sources=files,
                                        dependencies=deps)
    return self._inject_target(target, dependees, self.gen_python, 'py', create_target)

  def _inject_target(self, target, dependees, geninfo, namespace, create_target):
    files = []
    has_service = False
    for source in target.sources:
      services, genfiles = calculate_gen(os.path.join(target.target_base, source))
      has_service = has_service or services
      files.extend(genfiles.get(namespace, []))
    deps = geninfo.deps['service' if has_service else 'structs']
    tgt = create_target(files, deps)
    tgt.id = target.id
    tgt.is_codegen = True
    for dependee in dependees:
      if isinstance(dependee, InternalTarget):
        dependee.update_dependencies((tgt,))
      else:
        # TODO(John Sirois): rationalize targets with dependencies.
        # JarLibrary or PythonTarget dependee on the thrift target
        dependee.dependencies.add(tgt)
    return tgt


NAMESPACE_PARSER = re.compile(r'^\s*namespace\s+([^\s]+)\s+([^\s]+)\s*$')
TYPE_PARSER = re.compile(r'^\s*(const|enum|exception|service|struct|union)\s+([^\s{]+).*')


# TODO(John Sirois): consolidate thrift parsing to 1 pass instead of 2
def calculate_gen(source):
  """Calculates the service types and files generated for the given thrift IDL source.

  Returns a tuple of (service types, generated files).
  """

  with open(source, 'r') as thrift:
    lines = thrift.readlines()
    namespaces = {}
    types = defaultdict(set)
    for line in lines:
      match = NAMESPACE_PARSER.match(line)
      if match:
        lang = match.group(1)
        namespace = match.group(2)
        namespaces[lang] = namespace
      else:
        match = TYPE_PARSER.match(line)
        if match:
          typename = match.group(1)
          name = match.group(2)
          types[typename].add(name)

    genfiles = defaultdict(set)

    namespace = namespaces.get('py')
    if namespace:
      genfiles['py'].update(calculate_python_genfiles(namespace, types))

    namespace = namespaces.get('java')
    if namespace:
      genfiles['java'].update(calculate_java_genfiles(namespace, types))

    return types['service'], genfiles


def calculate_python_genfiles(namespace, types):
  basepath = namespace.replace('.', '/')
  def path(name):
    return os.path.join(basepath, '%s.py' % name)
  yield path('__init__')
  if 'const' in types:
    yield path('constants')
  if set(['enum', 'exception', 'struct', 'union']) & set(types.keys()):
    yield path('ttypes')
  for service in types['service']:
    yield path(service)
    yield os.path.join(basepath, '%s-remote' % service)


def calculate_java_genfiles(namespace, types):
  basepath = namespace.replace('.', '/')
  def path(name):
    return os.path.join(basepath, '%s.java' % name)
  if 'const' in types:
    yield path('Constants')
  for typename in ['enum', 'exception', 'service', 'struct', 'union']:
    for name in types[typename]:
      yield path(name)



from .console_task import ConsoleTask
from twitter.pants.targets import SourceRoot


class ListRoots(ConsoleTask):
  """
  List the registered source roots of the repo.
  """

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    super(ListRoots, cls).setup_parser(option_group, args, mkflag)

  @classmethod
  def console_output(self, targets):
    for srcroot, targets in SourceRoot.all_roots().items():
      all_targets = ",".join(sorted([tgt.__name__ for tgt in targets]))
      yield "%s: %s" % (srcroot, all_targets)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import hashlib
import logging
import os
import pkgutil
import re
import shutil
import time

from contextlib import contextmanager

from twitter.common.collections import OrderedDict, OrderedSet
from twitter.common.dirutil import safe_mkdir, safe_open
from twitter.common.log.options import LogOptions

from twitter.pants import binary_util, get_buildroot, is_internal, is_jar, is_jvm, is_concrete
from twitter.pants.base.generator import Generator, TemplateData
from twitter.pants.base.revision import Revision
from twitter.pants.ivy import Bootstrapper, Ivy
from twitter.pants.java import runjava

from .nailgun_task import NailgunTask

from . import TaskError


class IvyResolve(NailgunTask):

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    NailgunTask.setup_parser(option_group, args, mkflag)

    flag = mkflag('override')
    option_group.add_option(flag, action='append', dest='ivy_resolve_overrides',
                            help='''Specifies a jar dependency override in the form:
                            [org]#[name]=(revision|url)

                            For example, to specify 2 overrides:
                            %(flag)s=com.foo#bar=0.1.2 \\
                            %(flag)s=com.baz#spam=file:///tmp/spam.jar
                            ''' % dict(flag=flag))

    report = mkflag("report")
    option_group.add_option(report, mkflag("report", negate=True), dest = "ivy_resolve_report",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%default] Generate an ivy resolve html report")

    option_group.add_option(mkflag("open"), mkflag("open", negate=True),
                            dest="ivy_resolve_open", default=False,
                            action="callback", callback=mkflag.set_bool,
                            help="[%%default] Attempt to open the generated ivy resolve report "
                                 "in a browser (implies %s)." % report)

    option_group.add_option(mkflag("outdir"), dest="ivy_resolve_outdir",
                            help="Emit ivy report outputs in to this directory.")

    option_group.add_option(mkflag("args"), dest="ivy_args", action="append", default=[],
                            help = "Pass these extra args to ivy.")

    option_group.add_option(mkflag("mutable-pattern"), dest="ivy_mutable_pattern",
                            help="If specified, all artifact revisions matching this pattern will "
                                 "be treated as mutable unless a matching artifact explicitly "
                                 "marks mutable as False.")

  def __init__(self, context, confs=None):
    super(IvyResolve, self).__init__(context)

    self._ivy_bootstrapper = Bootstrapper.instance()
    self._cachedir = self._ivy_bootstrapper.ivy_cache_dir

    self._confs = confs or context.config.getlist('ivy-resolve', 'confs', default=['default'])
    self._transitive = context.config.getbool('ivy-resolve', 'transitive', default=True)

    self._jvm_args = context.config.getlist('ivy-resolve', 'jvm_args', default=[])
    # Disable cache in File.getCanonicalPath() to make Ivy work with -symlink option properly on
    # nailgun.
    self._jvm_args.append('-Dsun.io.useCanonCaches=false')

    self._ivy_args = (context.options.ivy_args or
                      context.config.getlist('ivy-resolve', 'ivy_args', default=[]))

    self._mutable_pattern = (context.options.ivy_mutable_pattern or
                             context.config.get('ivy-resolve', 'mutable_pattern', default=None))
    if self._mutable_pattern:
      try:
        self._mutable_pattern = re.compile(self._mutable_pattern)
      except re.error as e:
        raise TaskError('Invalid mutable pattern specified: %s %s' % (self._mutable_pattern, e))

    self._profile = context.config.get('ivy-resolve', 'profile')

    self._template_path = os.path.join('templates', 'ivy_resolve', 'ivy.mk')

    self._work_dir = context.config.get('ivy-resolve', 'workdir')
    self._classpath_file = os.path.join(self._work_dir, 'classpath')

    self._outdir = context.options.ivy_resolve_outdir or os.path.join(self._work_dir, 'reports')
    self._open = context.options.ivy_resolve_open
    self._report = self._open or context.options.ivy_resolve_report

    def parse_override(override):
      match = re.match(r'^([^#]+)#([^=]+)=([^\s]+)$', override)
      if not match:
        raise TaskError('Invalid dependency override: %s' % override)

      org, name, rev_or_url = match.groups()

      def fmt_message(message, template):
        return message % dict(
          overridden='%s#%s;%s' % (template.org, template.module, template.version),
          rev=rev_or_url,
          url=rev_or_url
        )

      def replace_rev(template):
        context.log.info(fmt_message('Overrode %(overridden)s with rev %(rev)s', template))
        return template.extend(version=rev_or_url, url=None, force=True)

      def replace_url(template):
        context.log.info(fmt_message('Overrode %(overridden)s with snapshot at %(url)s', template))
        return template.extend(version='SNAPSHOT', url=rev_or_url, force=True)

      replace = replace_url if re.match(r'^\w+://.+', rev_or_url) else replace_rev
      return (org, name), replace

    self._overrides = {}
    if context.options.ivy_resolve_overrides:
      self._overrides.update(parse_override(o) for o in context.options.ivy_resolve_overrides)

  def invalidate_for(self):
    return self.context.options.ivy_resolve_overrides

  def execute(self, targets):
    """
      Resolves the specified confs for the configured targets and returns an iterator over tuples
      of (conf, jar path).
    """

    def dirname_for_requested_targets(targets):
      """Where we put the classpath file for this set of targets."""
      sha = hashlib.sha1()
      for t in targets:
        sha.update(t.id)
      return sha.hexdigest()

    target_workdir = os.path.join(self._work_dir, dirname_for_requested_targets(targets))
    target_classpath_file = os.path.join(target_workdir, 'classpath')

    with self.changed(self._extract_classpathdeps(targets), only_buildfiles=True) as changed_deps:
      if changed_deps:
        self._exec_ivy(target_workdir, targets, [
          '-cachepath', target_classpath_file,
          '-confs'
        ] + self._confs)

    if os.path.exists(target_classpath_file):
      def safe_link(src, dest):
        if os.path.exists(dest):
          os.unlink(dest)
        os.symlink(src, dest)

      # Symlink to the current classpath file.
      safe_link(target_classpath_file, self._classpath_file)

      # Symlink to the current ivy.xml file (useful for IDEs that read it).
      ivyxml_symlink = os.path.join(self._work_dir, 'ivy.xml')
      target_ivyxml = os.path.join(target_workdir, 'ivy.xml')
      safe_link(target_ivyxml, ivyxml_symlink)

    with self._cachepath(self._classpath_file) as classpath:
      with self.context.state('classpath', []) as cp:
        for path in classpath:
          for conf in self._confs:
            cp.append((conf, path.strip()))

    if self._report:
      self._generate_ivy_report()

    create_jardeps_for = self.context.products.isrequired(self._mapfor_typename())
    if create_jardeps_for:
      genmap = self.context.products.get(self._mapfor_typename())
      for target in filter(create_jardeps_for, targets):
        self._mapjars(genmap, target)

  def _extract_classpathdeps(self, targets):
    """Subclasses can override to filter out a set of targets that should be resolved for classpath
    dependencies.
    """
    def is_classpath(target):
      return is_jar(target) or (
        is_internal(target) and any(jar for jar in target.jar_dependencies if jar.rev)
      )

    classpath_deps = OrderedSet()
    for target in targets:
      classpath_deps.update(filter(is_classpath, filter(is_concrete, target.resolve())))
    return classpath_deps

  def _generate_ivy(self, jars, excludes, ivyxml):
    org, name = self._identify()

    # We pair an <override/> with each forced dependency to effect transitive forces.
    dependencies = [self._generate_jar_template(jar) for jar in jars]
    overrides = [self._generate_override_template(dep) for dep in dependencies if dep.force]

    excludes = [self._generate_exclude_template(exclude) for exclude in excludes]

    template_data = TemplateData(
      org=org,
      module=name,
      version='latest.integration',
      publications=None,
      is_idl=False,
      dependencies=dependencies,
      excludes=excludes,
      overrides=overrides
    )

    safe_mkdir(os.path.dirname(ivyxml))
    with open(ivyxml, 'w') as output:
      generator = Generator(pkgutil.get_data(__name__, self._template_path),
                            root_dir = get_buildroot(),
                            lib = template_data)
      generator.write(output)

  def _generate_ivy_report(self):
    def make_empty_report(report, organisation, module, conf):
      no_deps_xml = """<?xml version="1.0" encoding="UTF-8"?>
<?xml-stylesheet type="text/xsl" href="ivy-report.xsl"?>
<ivy-report version="1.0">
	<info
		organisation="%(organisation)s"
		module="%(module)s"
		revision="latest.integration"
		conf="%(conf)s"
		confs="%(conf)s"
		date="%(timestamp)s"/>
</ivy-report>""" % dict(organisation=organisation,
                        module=module,
                        conf=conf,
                        timestamp=time.strftime('%Y%m%d%H%M%S'))
      with open(report, 'w') as report_handle:
        print(no_deps_xml, file=report_handle)

    classpath = self.profile_classpath(self._profile)

    reports = []
    org, name = self._identify()
    xsl = os.path.join(self._cachedir, 'ivy-report.xsl')

    # Xalan needs this dir to exist - ensure that, but do no more - we have no clue where this
    # points.
    safe_mkdir(self._outdir, clean=False)

    for conf in self._confs:
      params = dict(
        org=org,
        name=name,
        conf=conf
      )
      xml = os.path.join(self._cachedir, '%(org)s-%(name)s-%(conf)s.xml' % params)
      if not os.path.exists(xml):
        make_empty_report(xml, org, name, conf)
      out = os.path.join(self._outdir, '%(org)s-%(name)s-%(conf)s.html' % params)
      args = ['-IN', xml, '-XSL', xsl, '-OUT', out]
      xalan_main = 'org.apache.xalan.xslt.Process'
      # We use a shell-out here to save the nailgun for classpath resolves
      result = runjava(classpath, xalan_main, args=args)
      if result != 0:
        raise TaskError("java %s ... exited non-zero (%i)" % (xalan_main, result))
      reports.append(out)

    css = os.path.join(self._outdir, 'ivy-report.css')
    if os.path.exists(css):
      os.unlink(css)
    shutil.copy(os.path.join(self._cachedir, 'ivy-report.css'), self._outdir)

    if self._open:
      binary_util.ui_open(*reports)

  def _identify(self):
    if len(self.context.target_roots) == 1:
      target = self.context.target_roots[0]
      if hasattr(target, 'provides') and target.provides:
        return target.provides.org, target.provides.name
      else:
        return 'internal', target.id
    else:
      return 'internal', self.context.id

  def _calculate_classpath(self, targets):
    def is_jardependant(target):
      return is_jar(target) or is_jvm(target)

    jars = OrderedDict()
    excludes = set()
    # Support the ivy force concept when we sanely can for internal dep conflicts.
    # TODO(John Sirois): Consider supporting / implementing the configured ivy revision picking
    # strategy generally.
    def add_jar(jar):
      coordinate = (jar.org, jar.name)
      existing = jars.get(coordinate)
      jars[coordinate] = jar if not existing else (
        self._resolve_conflict(existing=existing, proposed=jar)
      )

    def collect_jars(target):
      if is_jar(target):
        add_jar(target)
      elif target.jar_dependencies:
        for jar in target.jar_dependencies:
          if jar.rev:
            add_jar(jar)

      # Lift jvm target-level excludes up to the global excludes set
      if is_jvm(target) and target.excludes:
        excludes.update(target.excludes)

    for target in targets:
      target.walk(collect_jars, is_jardependant)

    return jars.values(), excludes

  def _resolve_conflict(self, existing, proposed):
    if proposed == existing:
      return existing
    elif existing.force and proposed.force:
      raise TaskError('Cannot force %s#%s to both rev %s and %s' % (
        proposed.org, proposed.name, existing.rev, proposed.rev
      ))
    elif existing.force:
      self.context.log.debug('Ignoring rev %s for %s#%s already forced to %s' % (
        proposed.rev, proposed.org, proposed.name, existing.rev
      ))
      return existing
    elif proposed.force:
      self.context.log.debug('Forcing %s#%s from %s to %s' % (
        proposed.org, proposed.name, existing.rev, proposed.rev
      ))
      return proposed
    else:
      try:
        if Revision.lenient(proposed.rev) > Revision.lenient(existing.rev):
          self.context.log.debug('Upgrading %s#%s from rev %s  to %s' % (
            proposed.org, proposed.name, existing.rev, proposed.rev,
          ))
          return proposed
        else:
          return existing
      except Revision.BadRevision as e:
        raise TaskError('Failed to parse jar revision', e)

  def _is_mutable(self, jar):
    if jar.mutable is not None:
      return jar.mutable
    if self._mutable_pattern:
      return self._mutable_pattern.match(jar.rev)
    return False

  def _generate_jar_template(self, jar):
    template=TemplateData(
      org=jar.org,
      module=jar.name,
      version=jar.rev,
      mutable=self._is_mutable(jar),
      force=jar.force,
      excludes=[self._generate_exclude_template(exclude) for exclude in jar.excludes],
      transitive=jar.transitive,
      artifacts=jar.artifacts,
      is_idl='idl' in jar._configurations,
      configurations=';'.join(jar._configurations),
    )
    override = self._overrides.get((jar.org, jar.name))
    return override(template) if override else template

  def _generate_exclude_template(self, exclude):
    return TemplateData(org=exclude.org, name=exclude.name)

  def _generate_override_template(self, jar):
    return TemplateData(org=jar.org, module=jar.module, version=jar.version)

  @contextmanager
  def _cachepath(self, file):
    if not os.path.exists(file):
      yield ()
    else:
      with safe_open(file, 'r') as cp:
        yield (path.strip() for path in cp.read().split(os.pathsep) if path.strip())

  def _mapjars(self, genmap, target):
    mapdir = os.path.join(self._mapto_dir(), target.id)
    safe_mkdir(mapdir, clean=True)
    ivyargs = [
      '-retrieve', '%s/[organisation]/[artifact]/[conf]/'
                   '[organisation]-[artifact]-[revision](-[classifier]).[ext]' % mapdir,
      '-symlink',
      '-confs',
    ]
    ivyargs.extend(target.configurations or self._confs)
    self._exec_ivy(mapdir, [target], ivyargs)

    for org in os.listdir(mapdir):
      orgdir = os.path.join(mapdir, org)
      if os.path.isdir(orgdir):
        for name in os.listdir(orgdir):
          artifactdir = os.path.join(orgdir, name)
          if os.path.isdir(artifactdir):
            for conf in os.listdir(artifactdir):
              confdir = os.path.join(artifactdir, conf)
              for file in os.listdir(confdir):
                if self._map_jar(org, name, conf, file):
                  # TODO(John Sirois): kill the org and (org, name) exclude mappings in favor of a
                  # conf whitelist
                  genmap.add(org, confdir).append(file)
                  genmap.add((org, name), confdir).append(file)

                  genmap.add(target, confdir).append(file)
                  genmap.add((target, conf), confdir).append(file)
                  genmap.add((org, name, conf), confdir).append(file)

  def _mapfor_typename(self):
    """Subclasses can override to identify the product map typename that should trigger jar mapping.
    """
    return 'jar_dependencies'

  def _mapto_dir(self):
    """Subclasses can override to establish an isolated jar mapping directory."""
    return os.path.join(self._work_dir, 'mapped-jars')

  def _map_jar(self, org, name, conf, path):
    """Subclasses can override to determine whether a given path represents a mappable artifact."""
    return path.endswith('.jar')

  def _exec_ivy(self, target_workdir, targets, args):
    ivyxml = os.path.join(target_workdir, 'ivy.xml')
    jars, excludes = self._calculate_classpath(targets)
    self._generate_ivy(jars, excludes, ivyxml)

    ivy_args = ['-ivy', ivyxml]
    if LogOptions.stderr_log_level() == logging.DEBUG:
      ivy_args.append('-verbose')
    ivy_args.extend(args)
    if not self._transitive:
      ivy_args.append('-notransitive')
    ivy_args.extend(self._ivy_args)

    try:
      ivy = self._ivy_bootstrapper.ivy(self.java_executor)
      ivy.execute(ivy_args, jvm_args=self._jvm_args)
    except (Bootstrapper.Error, Ivy.Error) as e:
      raise TaskError('Failed to execute ivy call! %s' % e)



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from functools import partial

from twitter.common.collections import maybe_list

from twitter.pants.base import manual, TargetDefinitionException

from .internal import InternalTarget
from .jar_dependency import JarDependency
from .jar_library import JarLibrary
from .jarable import Jarable
from .pants_target import Pants
from .with_sources import TargetWithSources


@manual.builddict()
class ThriftLibrary(InternalTarget, TargetWithSources, Jarable):
  """Now-deprecated target that was used to generate code from thrift IDL files.
  You probably want some combination of ``java_thrift_target``,
  ``python_thrift_target``, ``*_thrift_target`` instead.
  """

  def __init__(self, name, sources, dependencies=None, provides=None):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param sources: A list of filenames representing the source code
      this library is compiled from.
    :type sources: list of strings
    :param dependencies: List of :class:`twitter.pants.base.target.Target` instances
      this target depends on.
    :type dependencies: list of targets
    :param provides: Broken optional :class:`twitter.pants.targets.Artifact` that describes how this
      thrift IDL library is published as a jar. (Alas, publishable
      targets need to depend on generated code, not just IDL. Thus,
      using ``provides`` probably doesn't do what you want.)
    """
    # TODO(John Sirois): Consider adding support for provides such that thrift-only IDL jar
    # publishing becomes an explicit choice pending inter-repo thrift-sharing idea shake-out.
    TargetWithSources.__init__(self, name, sources)
    InternalTarget.__init__(self, name,
                            maybe_list(dependencies or (),
                                       expected_type=(JarLibrary, Pants, ThriftJar, ThriftLibrary),
                                       raise_type=partial(TargetDefinitionException, self)))

    self._provides = provides
    self.is_codegen = True

  @property
  def provides(self):
    return self._provides

  def valid_dependency(self, dep):
    return isinstance(dep, (ThriftJar, ThriftLibrary))


@manual.builddict(tags=["anylang", "thrift"])
class ThriftJar(JarDependency):
  """Defines a jar that contains thrift IDL files that can be extracted and
  used to generate code.

  To ensure a consistent jar version across libraries, most repos wrap
  ThriftJar inside a :class:`twitter.pants.targets.jar_library.JarLibrary`
  in a a central location for 3rdparty library declarations. For example:

  ::

    jar_library(name='mybird-thrift',
      dependencies=[
        thrift_jar(org='com.twitter', name='mybird-thrift', rev='1.0.0')
      ]
    )
  """

  def __init__(self, org, name, rev=None, force=False, url=None, mutable=None, classifier='idl'):
    """
    :param string org: Equivalent to Maven groupId.
    :param string name: Equivalent to Maven artifactId.
    :param string rev: Equivalent to Maven version.
    :param boolean force: Force this version.
    :param string url: Artifact URL if using a non-standard repository layout.
    :param boolean mutable: If the artifact changes and should not be cached.
    :param string classifier: Equivalent to Maven classifier.
    """
    JarDependency.__init__(self, org, name, rev=rev, force=force, url=url, mutable=mutable)

    # TODO(John Sirois): abstract ivy specific configurations notion away
    self._configurations.append('idl')
    self.with_artifact(configuration='idl', classifier=classifier)



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import getpass
import socket

from collections import namedtuple
from time import strftime, localtime

from twitter.pants import get_buildroot, get_scm

BuildInfo = namedtuple('BuildInfo', 'date time timestamp branch tag sha user machine path')


def get_build_info(scm=None):
  """Calculates the current BuildInfo using the supplied scm or else the globally configured one
  if any.
  """
  now = localtime()
  buildroot = get_buildroot()

  scm = scm or get_scm()
  revision = scm.commit_id if scm else 'unknown'
  tag = (scm.tag_name or 'none') if scm else 'unknown'
  branchname = (scm.branch_name or revision) if scm else 'unknown'

  return BuildInfo(
      date=strftime('%A %b %d, %Y', now),
      time=strftime('%H:%M:%S', now),
      timestamp=strftime('%m.%d.%Y %H:%M', now),
      branch=branchname,
      tag=tag,
      sha=revision,
      user=getpass.getuser(),
      machine=socket.gethostname(),
      path=buildroot)



#!/usr/bin/env python2.7

import os
import os.path
import shutil

from string import Template


TEMPLATE = Template('\n'.join([
  ':mod:`$name` Module',
  '-----------------------------------------------',
  '',
  '.. automodule:: twitter.pants.$otype.$name',
  '   :members:',
  '', '',
]))

def gen_targets_reference(targets_rst, targets_dir):
  lines = [
    'Targets Reference',
    '=================',
    '',
    'This page documents targets available as part of the pants build system.',
    '', '',
  ]

  for filename in sorted([filename for filename in os.listdir(targets_dir) if filename.endswith('.py')]):
    if filename == '__init__.py':
      continue # Skip because renaming targets causes duplicates.
    root, _ = os.path.splitext(filename)
    lines.append(TEMPLATE.substitute(otype='targets', name=root))

  with open(targets_rst, 'w') as fh:
    fh.write('\n'.join(lines))

def gen_base_reference(rst_filename, dirname):
  lines = [
    'Base Reference',
    '==============',
    '',
    'This page documents base classes of the pants build system.',
    '', '',
  ]

  for filename in sorted([filename for filename in os.listdir(dirname) if filename.endswith('.py')]):
    if filename == '__init__.py':
      continue # Skip because renaming targets causes duplicates.
    root, _ = os.path.splitext(filename)
    lines.append(TEMPLATE.substitute(otype='base', name=root))

  with open(rst_filename, 'w') as fh:
    fh.write('\n'.join(lines))

def main():
  docs_dir = os.path.dirname(os.path.abspath(__file__))
  pants_src_dir = os.path.dirname(docs_dir)
  tasks_dir = os.path.join(pants_src_dir, 'tasks')

  for filename in ['build_dictionary.rst', 'goals_reference.rst']:
    filepath = os.path.abspath(os.path.join(pants_src_dir,
        '../../../../dist/builddict', filename))
    try:
      shutil.copy(filepath, docs_dir)
    except IOError as e:
      raise IOError("Forgot to `./pants goal builddict` first? \n\n%s" % e)

  with open(os.path.join(docs_dir, 'tasks.rst'), 'w') as tasks_rst:
    tasks_rst.write('\n'.join([
      'Tasks Reference',
      '===============',
      '',
      'This page documents tasks available as part of the pants build system.',
      '', '',
    ]))
    for filename in sorted([filename for filename in os.listdir(tasks_dir) if filename.endswith('.py')]):
      root, _ = os.path.splitext(filename)
      tasks_rst.write(TEMPLATE.substitute(otype='tasks', name=root))

  targets_rst = os.path.join(docs_dir, 'targets.rst')
  gen_targets_reference(targets_rst, os.path.join(pants_src_dir, 'targets'))

  gen_base_reference(os.path.join(docs_dir, 'base.rst'), os.path.join(pants_src_dir, 'base'))

if __name__ == '__main__':
  main()
